'''A REST API for the UniProt protein database and a parser for their XML format.
'''
import csv
import json
import warnings
import threading
import gzip
import io
import logging

from typing import List, Optional, ClassVar, Dict


try:
    from urllib import urlopen
except ImportError:
    from urllib.request import urlopen

from collections import defaultdict

import requests

from lxml import etree

from glypy.utils import make_struct
from glypy.io.glyspace import UniprotRDFClient

from glycopeptidepy.utils import simple_repr
from glycopeptidepy.structure.glycan import GlycosylationType

from six import add_metaclass, string_types as basestring

URI_TEMPLATE = "https://www.uniprot.org/uniprot/{accession}.xml"
BATCH_URI = "https://www.uniprot.org/uploadlists/"
NSMAP = {"up": "http://uniprot.org/uniprot"}
# BATCH_URI_QUERY = "https://rest.uniprot.org/uniprotkb/search?format=xml&query="
BATCH_URI_QUERY = "https://rest.uniprot.org/uniprotkb/accessions?format=xml&accessions="

VERIFY_SSL = False

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _batch_uri_query_builder(accessions: List[str]) -> str:
    # query = " OR ".join([f"accession:{acc}" for acc in accessions])
    query = ','.join(accessions)
    return BATCH_URI_QUERY + query


class UniProtFeatureMeta(type):
    _cache: ClassVar[Dict[str, 'UniProtFeatureMeta']] = {}

    def __new__(cls, name, parents, attrs):
        new_type = type.__new__(cls, name, parents, attrs)
        try:
            cls._cache[new_type.feature_type] = new_type
        except AttributeError:
            pass
        return new_type

    def from_feature_type(self, feature_type: str) -> 'UniProtFeatureMeta':
        return self._cache[feature_type]

    def handle_tag(self, tag: etree.Element) -> 'UniProtFeatureBase':
        feature_type = tag.attrib['type']
        try:
            feature_type_t = self.from_feature_type(feature_type)  # pylint: disable=no-value-for-parameter
            return feature_type_t.fromxml(tag)
        except KeyError:
            return None


@add_metaclass(UniProtFeatureMeta)
class UniProtFeatureBase(object):
    __repr__ = simple_repr

    known: bool = True
    _name: Optional[str] = None

    @property
    def is_defined(self):
        return self.known

    def describe(self) -> str:
        return self.description.split("; ")

    @property
    def name(self):
        return self.feature_type


class Keyword(str):
    id: str

    def __new__(cls, content, idstr):
        obj = str.__new__(cls, content)
        obj.id = idstr
        return obj

    def __reduce__(self):
        return self.__class__, (str(self), self.id)


class PeptideBase(UniProtFeatureBase):
    start: int
    end: int

    def __init__(self, start, end, known=True):
        self.start = start
        self.end = end
        self.known = known

    @classmethod
    def fromxml(cls, feature: etree.Element):
        known = True

        try:
            begin = int(feature.find(".//{http://uniprot.org/uniprot}begin", NSMAP).attrib['position']) - 1
        except (KeyError, AttributeError):
            begin = 0
            known = False
        try:
            end = int(feature.find(".//{http://uniprot.org/uniprot}end", NSMAP).attrib['position'])
        except (KeyError, AttributeError):
            end = float('inf')
            known = False
        if not known:
            if begin == 0 and end == float('inf'):
                begin = int(feature.find(
                    ".//{http://uniprot.org/uniprot}location", NSMAP).attrib['position']) - 1
                end = begin + 1
                known = True
        return cls(begin, end, known)


class SignalPeptide(PeptideBase):
    feature_type = 'signal peptide'


class Propeptide(PeptideBase):
    feature_type = 'propeptide'


class TransitPeptide(PeptideBase):
    feature_type = "transit peptide"


class Peptide(PeptideBase):
    feature_type = 'peptide'


class MatureProtein(PeptideBase):
    feature_type = 'mature protein'


class SpanBase(UniProtFeatureBase):
    feature_type = "__spanbase__"

    start: int
    end: int

    def __init__(self, name, start, end, known=True):
        self.name = name
        self.start = start
        self.end = end
        self.known = known

    @classmethod
    def fromxml(cls, feature: etree.Element):
        known = True
        try:
            try:
                begin = int(feature.find(
                    ".//up:begin", NSMAP).attrib['position']) - 1
            except KeyError:
                begin = 0
                known = False
            try:
                end = int(feature.find(
                    ".//up:end", NSMAP).attrib['position'])
            except KeyError:
                end = float('inf')
                known = False
        except AttributeError:
            begin = int(feature.find(
                ".//up:position", NSMAP).attrib['position']) - 1
            end = begin + 1

        description = feature.attrib.get("description")
        return cls(description, begin, end, known)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self.name

    def __repr__(self):
        repr_str = super(SpanBase, self).__repr__()
        prefix, suffix = repr_str.split("(", 1)
        augmented = "%s(name=%r, %s" % (prefix, self.name, suffix)
        return augmented


class Domain(SpanBase):
    feature_type = 'domain'


class Chain(SpanBase):
    feature_type = 'chain'


class RegionOfInterest(Domain):
    feature_type = 'region of interest'


class InitiatorMethionine(Domain):
    feature_type = "initiator methionine"


class DisulfideBond(Domain):
    feature_type = "disulfide bond"

    @classmethod
    def fromxml(cls, feature):
        name = "disulfide bond"
        known = True
        try:
            try:
                begin = int(feature.find(
                    ".//up:begin", NSMAP).attrib['position']) - 1
            except KeyError:
                begin = 0
                known = False
            try:
                end = int(feature.find(
                    ".//up:end", NSMAP).attrib['position'])
            except KeyError:
                end = float('inf')
                known = False
        except AttributeError:
            begin = end = int(feature.find(
                ".//up:position", NSMAP).attrib['position'])
        return cls(name, begin, end, known)


class ModifiedResidue(UniProtFeatureBase):
    feature_type = 'modified residue'

    position: int
    modification: str

    def __init__(self, position, modification):
        self.position = position
        self.modification = modification

    @property
    def name(self):
        return self.modification

    @property
    def description(self):
        return self.modification

    @property
    def start(self):
        return self.position

    @property
    def end(self):
        return self.position + 1

    @classmethod
    def fromxml(cls, feature):
        return cls(
            int(feature.find(
                ".//up:position", NSMAP).attrib['position']) - 1,
            feature.attrib["description"])


class SequenceVariant(UniProtFeatureBase):
    feature_type = "sequence variant"

    position: int
    original: str
    variation: str

    def __init__(self, position, original, variation):
        self.position = position
        self.original = original
        self.variation = variation

    @property
    def start(self):
        return self.position

    @property
    def end(self):
        return self.position + 1

    @classmethod
    def fromxml(cls, feature):
        position = feature.find(".//up:position", NSMAP)
        if position is None:
            # Deletion mutations not supported
            return None
        else:
            position = int(position.attrib['position']) - 1
            original = feature.find(".//up:original", NSMAP)
            if original is not None:
                original = original.text
            variation = feature.find(".//up:variation", NSMAP)
            if variation is not None:
                variation = variation.text
        return cls(position, original, variation)

    @property
    def description(self):
        return "%s->%s" % (self.original, self.variation)


class Site(Domain):
    feature_type = 'site'


class GlycosylationSite(UniProtFeatureBase):
    feature_type = 'glycosylation site'

    position: int
    glycosylation_type: str
    description: str

    def __init__(self, position, glycosylation_type, description=None):
        self.position = position
        try:
            self.glycosylation_type = GlycosylationType[glycosylation_type]
        except KeyError:
            self.glycosylation_type = GlycosylationType[None]
        self.description = description

    @property
    def name(self):
        return self.glycosylation_type

    @property
    def start(self):
        return self.position

    @property
    def end(self):
        return self.position + 1

    @classmethod
    def fromxml(cls, feature):
        position = int(feature.find(
            ".//up:position", NSMAP).attrib['position']) - 1
        glycosylation_type = feature.attrib["description"].split(" ")[0]
        return cls(position, glycosylation_type, feature.attrib["description"])



# @dataclass
# class UniProtProtein:
#     sequence: str
#     features: List[UniProtFeatureBase]
#     recommended_name: str
#     gene_name: str
#     names: List[str]
#     accessions: List[str]
#     keywords: List[Keyword]
#     dbreferences: List


_UniProtProtein = make_struct("UniProtProtein", (
    "sequence", "features", "recommended_name", "gene_name", "names", "accessions",
    "keywords", "dbreferences"))


class UniProtProtein(_UniProtProtein):
    __slots__ = ()

    def features_of_type(self, feature_type):
        if isinstance(feature_type, basestring):
            return [feature for feature in self.features if feature.feature_type == feature_type]
        elif isinstance(feature_type, type):
            return [feature for feature in self.features if isinstance(feature, feature_type)]
        else:
            raise TypeError("Can't use object of type %r, %r" % (type(feature_type), feature_type))

    @property
    def chains(self):
        return self.features_of_type(Chain)

    @property
    def domains(self):
        return self.features_of_type(Domain)

    @property
    def variants(self):
        return self.features_of_type(SequenceVariant)

    @property
    def peptides(self):
        return self.features_of_type(PeptideBase)


def get_etree_for(accession):
    tree = etree.parse(_open_url_for(accession))
    return tree


def _open_url_for(accession):
    return urlopen(URI_TEMPLATE.format(accession=accession))


def parse(tree, error=False):
    '''Parse an XML document into a :class:`UniProtProtein`.

    Parameters
    ----------
    tree: lxml.etree.Element
        The root XML element of the tree to parse.
    errors: bool
        Whether to treat errors during feature parsing
        as warnings or exceptions.

    Returns
    -------
    UniProtProtein
    '''
    entry = tree.find(".//up:entry", NSMAP)
    if entry is None:
        if hasattr(tree, "tag") and tree.tag.endswith("entry"):
            entry = tree
        else:
            raise ValueError("Could not find root entry!")
    seq = entry.find(
        "./up:sequence", NSMAP).text
    if seq is not None:
        seq = seq.replace("\n", '')
    names = [el.text for el in entry.findall(
        ".//up:protein/*/up:fullName", NSMAP)] + [el.text for el in entry.findall(
            ".//up:protein/*/up:shortName", NSMAP)]
    recommended_name_tags = entry.findall(
        ".//up:protein/up:recommendedName", NSMAP) + entry.findall(".//up:protein/*/up:recommendedName", NSMAP)
    if recommended_name_tags:
        recommended_name_tag = recommended_name_tags[0]
    else:
        recommended_name_tag = None
    if recommended_name_tag is not None:
        if recommended_name_tag.text.strip():
            recommended_name = recommended_name_tag.text.strip()
        else:
            recommended_name = (recommended_name_tag[0].text)
    else:
        try:
            recommended_name = names[0]
        except IndexError:
            recommended_name = ""
    gene_name_tag = entry.find(".//up:name", NSMAP)
    if gene_name_tag is not None:
        gene_name = gene_name_tag.text
    else:
        gene_name = ""
    accessions = [el.text for el in entry.findall(
        ".//up:accession", NSMAP)]
    dbreferences = defaultdict(list)
    for tag in entry.findall(".//up:dbReference", NSMAP):
        db = tag.attrib['type']
        rec = dict(id=tag.attrib['id'])
        for prop in tag.findall(".//up:property", NSMAP):
            rec[prop.attrib['type']] = prop.attrib['value']
        dbreferences[db].append(rec)

    features = []
    exc_type = Exception
    for tag in entry.findall(".//up:feature", NSMAP):
        feature_type = tag.attrib['type']
        try:
            feature_obj = UniProtFeatureBase.handle_tag(tag)
            if feature_obj is not None:
                features.append(feature_obj)
        except exc_type as e:
            if error:
                raise
            else:
                warnings.warn("An exception %r occurred while parsing feature type %s for %s" % (
                    e, feature_type, accessions[0]))
    keywords = set()
    for kw in entry.findall(".//up:keyword", NSMAP):
        keywords.add(Keyword(kw.text, kw.attrib['id']))
    return UniProtProtein(
        seq, features, recommended_name, gene_name, names, accessions,
        keywords, dbreferences)


def parse_all(tree, error=False):
    """Parse an XML document containing multiple <up:entry> elements into
    a list of :class:`UniProtProtein` instances.

    Parameters
    ----------
    tree : lxml.etree.Element
        The root element of the XML document to parse.
    errors: bool
        Whether to treat errors during feature parsing
        as warnings or exceptions.

    Returns
    -------
    list
    """
    entries = tree.findall(".//up:entry", NSMAP)
    return [parse(e, error=error) for e in entries]


def iterparse(stream, error=False):
    for event, tag in etree.iterparse(stream, huge_tree=True, events=("start", "end")):
        if event == "end":
            if tag.tag.split("}")[1] == 'entry':
                yield parse(tag, error=error)


def get_features_for(accession, error=False):
    tree = get_etree_for(accession)
    return parse(tree, error)


def get_features_for_many(accessions, error=False):
    query_to_doc = dict()
    url = _batch_uri_query_builder(accessions)
    r = requests.get(url, verify=VERIFY_SSL)
    tree = etree.fromstring(r.content)
    for doc in parse_all(tree, error=error):
        for acc in doc.accessions:
            query_to_doc[acc] = doc
    results = []
    for acc in accessions:
        if acc in query_to_doc:
            results.append((acc, query_to_doc[acc]))
        else:
            results.append((acc, None))
    return results


def get(accessions):
    if accessions is None:
        raise TypeError("accessions cannot be `None`")
    if isinstance(accessions, basestring):
        return get_features_for(accessions)
    elif len(accessions) < 5:
        return [b for a, b in get_features_for_many(accessions) if b is not None]
    else:
        chunk_size = min(len(accessions) // 5, 15)
        return ProteinDownloader.download(accessions, chunk_size)


class ProteinDownloader(object):  # pragma: no cover
    @staticmethod
    def chunk(seq, n=5):
        i = 0
        while i < len(seq):
            yield seq[i:i + n]
            i += n

    @staticmethod
    def fetch(chunk):
        jobs = []
        results = []

        def task(name):
            try:
                feat = get_features_for(name)
                results.append(feat)
            except Exception as e:
                print(e, name)

        for c in chunk:
            th = threading.Thread(target=task, args=(c,))
            th.start()
            jobs.append(th)

        for th in jobs:
            th.join()

        return results

    @classmethod
    def download(cls, accession_iterable, chunk_size=15, ordered=True):
        accession_iterable = list(accession_iterable)
        chunks = cls.chunk(accession_iterable, chunk_size)
        proteins = []
        for chunk in chunks:
            proteins.extend(cls.fetch(chunk))
        if not ordered:
            return proteins
        acc_map = {}
        for prot in proteins:
            for acc in prot.accessions:
                acc_map[acc] = prot
        proteins = []
        for acc in accession_iterable:
            proteins.append(acc_map.get(acc))
        return proteins

    def __call__(self, *args, **kwargs):
        return self.download(*args, **kwargs)

    @classmethod
    def to_fasta_file(cls, accession_iterable, file_handle, chunk_size=15):
        from .fasta import FastaFileWriter
        proteins = cls.download(accession_iterable, chunk_size=chunk_size)
        writer = FastaFileWriter(file_handle)
        for protein in proteins:
            defline = ">sp|%s|%s %s" % (protein.accessions[0], protein.gene_name, protein.recommended_name)
            defline = defline.replace("\n", " ")
            seq = protein.sequence
            writer.write(defline, seq)
        return file_handle


rdf = UniprotRDFClient()


def search(query):
    url = (
        r"https://rest.uniprot.org/uniprot/search?"
        f"compressed=false&query={query}&format=tsv"
        r"&fields=id,accession,protein_name,organism_name,reviewed,gene_names,annotation_score,length")
    response = requests.get(url, stream=True, verify=VERIFY_SSL)
    response.raise_for_status()
    response_buffer = response.raw
    data = response_buffer.read()
    if response.headers.get("content-encoding") == 'gzip':
        response_buffer = gzip.GzipFile(mode='rb', fileobj=io.BytesIO(data))
        data = response_buffer.read().decode('utf-8')
    else:
        data = data.decode("utf-8")
    reader = csv.reader(data.splitlines(), delimiter='\t')
    header = next(reader)
    return [dict(zip(header, line)) for line in reader]


class _UniProtRestClientBase(object):

    def _buffer_to_tsv(self, data: bytes) -> csv.DictReader:
        return csv.DictReader(io.StringIO(data.decode("utf8")), delimiter='\t')

    def _buffer_to_json(self, data: bytes) -> dict:
        return json.loads(data.decode("utf8"))

    def _get_stream(self, url: str) -> io.IOBase:
        response = requests.get(url, stream=True, verify=VERIFY_SSL)
        response.raise_for_status()
        response_buffer = response.raw
        if response.headers.get("content-encoding") == 'gzip':
            response_buffer = gzip.GzipFile(mode='rb', fileobj=response_buffer)
        return response_buffer

    def _get_response_buffer(self, url: str) -> bytes:
        response = requests.get(url, stream=True, verify=VERIFY_SSL)
        response.raise_for_status()
        response_buffer = response.raw
        data = response_buffer.read()
        if response.headers.get("content-encoding") == 'gzip':
            response_buffer = gzip.GzipFile(
                mode='rb', fileobj=io.BytesIO(data))
            data = response_buffer.read()
        return data



class ProteomeSearchResult(dict):
    bind = None

    def download(self, raw=False, format='fasta'):
        return self.bind.get(self['Proteome ID'], raw=raw, format=format)


class ProteomeClient(_UniProtRestClientBase):

    SearchResultType = ProteomeSearchResult

    def search(self, format='tsv', raw=False, query=None, **kwargs):
        if query is None:
            query = ""
        query += " AND ".join(["%s:%s" % (k, v) for k, v in kwargs.items()])
        uri = "https://rest.uniprot.org/proteomes/search?query={query}&format={format}"
        url = uri.format(query=query, format=format)
        data = self._get_response_buffer(url)
        if raw:
            return io.BytesIO(data)
        if format == 'tsv':
            result = list(map(self.SearchResultType, self._buffer_to_tsv(data)))
            for item in result:
                item.bind = self
            return result
        elif format == 'json':
            return json.loads(data.decode("utf8"))
        elif format == 'list':
            result = []
            for line in data.splitlines():
                r = self.SearchResultType({
                    "Proteome ID": line.strip()
                })
                r.bind = self
                result.append(r)
            return result
        return io.BytesIO(data)

    def get(self, proteome_id, include_isoforms=False, raw=False, format='fasta', stream: bool=False):
        uri = "https://rest.uniprot.org/uniprotkb/stream?{query}&{params}&format={format}"
        query = {"proteome": proteome_id, }
        query = '&'.join(["query=%s:\"%s\"" % (k, v) for k, v in query.items()])
        params = {"includeIsoform": str(bool(include_isoforms)).lower()}
        params = '&'.join(["%s=%s" % (k, v) for k, v in params.items()])
        url = uri.format(query=query, params=params, format=format)
        if stream:
            stream = self._get_stream(url)
            if format == "fasta":
                from .fasta import ProteinFastaFileReader
                return ProteinFastaFileReader(stream, index=False)
            return stream
        data = self._get_response_buffer(url)
        if raw:
            return io.BytesIO(data)
        if format == 'fasta':
            from .fasta import ProteinFastaFileReader
            return ProteinFastaFileReader(io.BytesIO(data))
        elif format == 'xml':
            return etree.fromstring(data)
        else:
            return io.BytesIO(data)

proteome = ProteomeClient()


class TaxonomyClient(_UniProtRestClientBase):

    def search(self, format='tsv', raw=False, query=None, params=None, **kwargs):
        if query is None:
            query = ''
        if params is None:
            params = dict()
        query += "AND".join(["%s:%s" % (k, v)
                                      for k, v in kwargs.items()])
        params = '&'.join(["%s=%s" % (k, v) for k, v in params.items()])
        if params:
            params += '&'
        uri = "https://rest.uniprot.org/taxonomy/search?query={query}&{params}format={format}"
        url = uri.format(query=query, params=params, format=format)
        data = self._get_response_buffer(url)
        if raw:
            return io.BytesIO(data)
        if format == 'tsv':
            result = list(self._buffer_to_tsv(data))
            return result
        if format == 'json':
            return json.loads(data.decode('utf8'))
        elif format == 'list':
            result = data.splitlines()
            return result
        return io.BytesIO(data)


taxonomy = TaxonomyClient()



import io
import json
import zlib
import warnings

from typing import List, Sequence, Union, TextIO

from six import add_metaclass

from glycopeptidepy.structure.glycan import GlycosylationType

from . import uniprot
from . import fasta


class AnnotationMeta(type):
    _cache = {}
    _is_cleavable = set()
    _is_cleavable_tags = set()

    def __new__(mcs, name, parents, attrs):
        new_type = type.__new__(mcs, name, parents, attrs)
        try:
            mcs._cache[name] = new_type
        except AttributeError:
            pass
        if new_type.cleavable and getattr(new_type, 'feature_type', None):
            mcs._is_cleavable.add(new_type)
            mcs._is_cleavable_tags.add(new_type.feature_type)
        return new_type

    def _type_for_name(cls, feature_type: str):
        return cls._cache[feature_type]

    def from_dict(cls, d: dict) -> 'AnnotationBase':
        name = d.pop('__class__')
        impl = cls._type_for_name(name)  # pylint: disable=no-value-for-parameter
        d.pop("feature_type", None)
        return impl._coerce_state(d)

    def _coerce_state(cls, state: dict):
        return cls(**state)

    @property
    def cleavable_types(cls):
        return cls._is_cleavable | cls._is_cleavable_tags


@add_metaclass(AnnotationMeta)
class AnnotationBase(object):
    feature_type: str
    description: str
    cleavable: bool = False

    def __init__(self, feature_type, description, **kwargs):
        self.feature_type = feature_type
        self.description = description

    def to_dict(self):
        d = {}
        d['feature_type'] = self.feature_type
        d['description'] = self.description
        d['__class__'] = self.__class__.__name__
        return d

    def __eq__(self, other: 'AnnotationBase'):
        if other is None:
            return False
        return self.feature_type == other.feature_type and self.description == other.description

    def __ne__(self, other: 'AnnotationBase'):
        return not (self == other)

    def __hash__(self):
        return hash((self.feature_type, self.description))


class AnnotatedResidue(AnnotationBase):
    position: int

    def __init__(self, position, feature_type, description, **kwargs):
        super(AnnotatedResidue, self).__init__(feature_type, description, **kwargs)
        self.position = position

    @property
    def start(self):
        return self.position

    @property
    def end(self):
        return self.position + 1

    def to_dict(self):
        d = super(AnnotatedResidue, self).to_dict()
        d['position'] = self.position
        return d

    def __eq__(self, other):
        res = super(AnnotatedResidue, self).__eq__(other)
        if not res:
            return res
        return self.position == other.position

    def __hash__(self):
        return hash((self.feature_type, self.description, self.position))


class AnnotatedInterval(AnnotationBase):
    start: int
    end: int

    def __init__(self, start, end, feature_type, description, **kwargs):
        super(AnnotatedInterval, self).__init__(
            feature_type, description, **kwargs)
        self.start = start
        self.end = end

    def to_dict(self):
        d = super(AnnotatedInterval, self).to_dict()
        d['start'] = self.start
        d['end'] = self.end
        return d

    def __eq__(self, other):
        res = super(AnnotatedInterval, self).__eq__(other)
        if not res:
            return res
        return self.start == other.start and self.end == other.end

    def __hash__(self):
        return hash((self.feature_type, self.description, self.start, self.end))


class Domain(AnnotatedInterval):
    feature_type = 'domain'

    domain_type: str

    def __init__(self, domain_type, start, end, **kwargs):
        super(Domain, self).__init__(
            start, end, self.feature_type, self.feature_type, **kwargs)
        self.domain_type = domain_type

    @property
    def name(self):
        return self.domain_type

    def __repr__(self):
        template = '{self.__class__.__name__}({self.domain_type!r}, {self.start}, {self.end})'
        return template.format(self=self)

    def to_dict(self):
        d = super(Domain, self).to_dict()
        d['domain_type'] = self.domain_type
        return d


class PeptideBase(AnnotatedInterval):
    feature_type = None
    cleavable: bool = True

    def __init__(self, start, end, **kwargs):
        super(PeptideBase, self).__init__(
            start, end, self.feature_type, self.feature_type, **kwargs)

    def __repr__(self):
        template = '{self.__class__.__name__}({self.start}, {self.end})'
        return template.format(self=self)

    def to_dict(self):
        state = super().to_dict()
        state.pop("description")
        return state


class SignalPeptide(PeptideBase):
    feature_type = 'signal peptide'


class Propeptide(PeptideBase):
    feature_type = 'propeptide'


class TransitPeptide(PeptideBase):
    feature_type = "transit peptide"


class Peptide(PeptideBase):
    feature_type = 'peptide'


class MaturationPeptide(PeptideBase):
    feature_type = 'maturation peptide'


class MatureProtein(PeptideBase):
    feature_type = 'mature protein'


class ProteolyticSite(AnnotatedResidue):
    feature_type = 'proteolytic site'

    cleavable: bool = True

    def __init__(self, position, **kwargs):
        super(ProteolyticSite, self).__init__(
            position,
            self.feature_type,
            self.feature_type,
            **kwargs)


class ModifiedResidue(AnnotatedResidue):
    feature_type = 'modified residue'

    def __init__(self, position, modification, **kwargs):
        super(ModifiedResidue, self).__init__(
            position, self.feature_type, modification, **kwargs)

    @property
    def modification(self):
        return self.description

    def to_dict(self):
        d = super(ModifiedResidue, self).to_dict()
        d['modification'] = d.pop('description')
        return d

    def __repr__(self):
        return "{self.__class__.__name__}({self.position}, {self.modification!r})".format(self=self)


class GlycosylationSite(AnnotatedResidue):
    feature_type = 'glycosylation site'

    glycosylation_type: GlycosylationType

    def __init__(self, position, glycosylation_type, **kwargs):
        if isinstance(glycosylation_type, str):
            glycosylation_type = GlycosylationType[glycosylation_type]
        super().__init__(position, self.feature_type, glycosylation_type, **kwargs)
        self.glycosylation_type = glycosylation_type

    def to_dict(self):
        d = super().to_dict()
        d['glycosylation_type'] = d.pop('description').name
        return d

    def __repr__(self):
        return "{self.__class__.__name__}({self.position}, {self.glycosylation_type!r})".format(self=self)


class SimpleVariant(AnnotatedResidue):
    feature_type = "simple variant"

    def __init__(self, position, substitution, **kwargs):
        super(SimpleVariant, self).__init__(position, self.feature_type, substitution, **kwargs)

    @property
    def substitution(self):
        return self.description

    def to_dict(self):
        d = super(SimpleVariant, self).to_dict()
        d['substitution'] = d.pop('description')
        return d


class ComplexVariant(AnnotatedInterval):
    feature_type = 'complex variant'

    def __init__(self, start, end, substitution, **kwargs):
        super(ComplexVariant, self).__init__(
            start, end, self.feature_type, substitution, **kwargs)

    @property
    def substitution(self):
        return self.description

    def to_dict(self):
        d = super(ComplexVariant, self).to_dict()
        d['substitution'] = d.pop('description')
        return d


class AnnotationCollection(Sequence[Union[AnnotationBase, AnnotatedInterval, AnnotatedResidue]]):
    items: List[Union[AnnotatedResidue, AnnotatedInterval, AnnotationBase]]

    def cleavable(self):
        return sorted({item for item in self if item.cleavable}, key=lambda x: (x.start, x.end))

    def modifications(self):
        return sorted({item for item in self if isinstance(item, (ModifiedResidue, GlycosylationSite))},
                      key=lambda x: x.start)

    def __init__(self, items):
        self.items = list(items)

    def append(self, item: AnnotationBase):
        self.items.append(item)

    def __getitem__(self, i) -> AnnotationBase:
        return self.items[i]

    def __setitem__(self, i, item):
        self.items[i] = item

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def __repr__(self):
        return "{self.__class__.__name__}({self.items})".format(self=self)

    def to_dict(self):
        return [a.to_dict() for a in self]

    @classmethod
    def from_dict(cls, d):
        return cls([AnnotationBase.from_dict(di) for di in d])

    def dump(self, fp: TextIO):
        json.dump(self.to_dict(), fp)

    @classmethod
    def load(cls, fp: TextIO):
        d = json.load(fp)
        inst = cls.from_dict(d)
        return inst

    def to_bytes(self) -> bytes:
        buffer = io.StringIO()
        self.dump(buffer)
        return zlib.compress(buffer.getvalue().encode("utf8"))

    @classmethod
    def from_bytes(cls, data: bytes) -> 'AnnotationCollection':
        data = zlib.decompress(data)
        return cls.load(io.StringIO(data.decode("utf8")))

    def __reduce__(self):
        return self.from_bytes, (self.to_bytes(), )

    def __eq__(self, other):
        return self.items == list(other)

    def __ne__(self, other):
        return not (self == other)


def from_uniprot(record: uniprot.UniProtProtein) -> AnnotationCollection:
    mapping = {
        uniprot.SignalPeptide.feature_type: SignalPeptide,
        uniprot.Propeptide.feature_type: Propeptide,
        uniprot.TransitPeptide.feature_type: TransitPeptide,
        uniprot.MatureProtein.feature_type: MatureProtein,
        uniprot.ModifiedResidue.feature_type: ModifiedResidue,
        uniprot.GlycosylationSite: GlycosylationSite,
    }
    annotations = []
    for feature in record.features:
        if not feature.is_defined:
            continue
        try:
            annotation_tp = mapping[feature.feature_type]
            if issubclass(annotation_tp, PeptideBase):
                annotations.append(
                    annotation_tp(feature.start, feature.end))
            elif isinstance(annotation_tp, GlycosylationSite):
                annotations.append(
                    annotation_tp(feature.position, feature.glycosylation_type))
            elif issubclass(annotation_tp, AnnotatedResidue):
                annotations.append(
                    annotation_tp(feature.position, feature.description))
        except KeyError:
            continue
    return AnnotationCollection(annotations)


def from_peff(record: fasta.PEFFFastaHeader) -> AnnotationCollection:
    annotations = []
    start: int
    end: int
    name: str
    accession: str
    node: tuple

    for node in record.get("Processed", []):
        if len(node) == 3:
            (start, end, name) = node
        else:
            (start, end, accession, name) = node
        start = start - 1
        if name == "signal peptide":
            annotations.append(
                SignalPeptide(start, end)
            )
        elif name == "mature protein":
            annotations.append(
                MatureProtein(start, end)
            )
        elif name == "transit peptide":
            annotations.append(
                TransitPeptide(start, end)
            )
        elif name == "propeptide":
            annotations.append(
                Propeptide(start, end)
            )
        elif name == "maturation peptide":
            annotations.append(
                MaturationPeptide(start, end)
            )
        else:
            warnings.warn(f"Unknown processing annotation {start}-{end} {name}")

    mods = record.get("ModRes", []) + record.get("ModResUnimod", [])
    for position, accession, name in mods:
        position = position - 1
        if accession:
            annotations.append(
                ModifiedResidue(position, accession))
        else:
            tokens = name.split(" ")
            if tokens[0] in GlycosylationType:
                glycosylation_type = GlycosylationType[tokens[0]]
                if glycosylation_type == GlycosylationType.o_linked:
                    if "glycosaminoglycan" in name:
                        glycosylation_type = GlycosylationType.glycosaminoglycan
                annotations.append(
                    GlycosylationSite(position, glycosylation_type))
    return AnnotationCollection(annotations)




from contextlib import contextmanager
from collections import defaultdict


def identity(x):
    return x


class TreeReprMixin(object):
    def __repr__(self):
        base = dict(self)
        return repr(base)


class PrefixTree(TreeReprMixin, defaultdict):
    '''
    A hash-based Prefix Tree for testing for
    sequence inclusion. This implementation works for any
    slice-able sequence of hashable objects, not just strings.
    '''
    def __init__(self):
        defaultdict.__init__(self, PrefixTree)
        self.labels = set()

    def add(self, sequence, label=None):
        """Adds a single Sequence-like object to the tree
        structure, placing the ith member of `sequence` on
        the ith level of `self`. At each level, if `label`
        is not None, it is added to the level's `labels` set.

        Parameters
        ----------
        sequence : Sequence
            Sequence-like object to add to the tree
        label : object, optional
            An arbitrary object which denotes or is associated with
            `sequence`

        Returns
        -------
        self
        """
        layer = self
        if label is None:
            label = sequence
        if label:
            layer.labels.add(label)
        for i in range(len(sequence)):
            layer = layer[sequence[i]]
            if label:
                layer.labels.add(label)

        return self

    def add_ngram(self, sequence, label=None):
        """Adds successive prefixes of `sequence`
        to the tree by calling :meth:`add`. This process
        is identical to calling :meth:`add` once for a prefix
        tree.

        Parameters
        ----------
        sequence : Sequence
            Sequence-like object to add to the tree
        label : object, optional
            An arbitrary object which denotes or is associated with
            `sequence`
        """
        if label is None:
            label = sequence
        for i in range(1, len(sequence) + 1):
            self.add(sequence[:i], label)

    def __contains__(self, sequence):
        layer = self
        j = 0
        for i in sequence:
            if not dict.__contains__(layer, i):
                break
            layer = layer[i]
            j += 1
        return len(sequence) == j

    def depth_in(self, sequence):
        layer = self
        count = 0
        for i in sequence:
            if not dict.__contains__(layer, i):
                break
            else:
                layer = layer[i]
            count += 1
        return count

    def subsequences_of(self, sequence):
        layer = self
        for i in sequence:
            layer = layer[i]
        return layer.labels

    def __iter__(self):
        return iter(self.labels)


class SuffixTree(PrefixTree):
    '''
    A hash-based Suffix Tree for testing for
    sequence inclusion. This implementation works for any
    slice-able sequence of hashable objects, not just strings.
    '''
    def __init__(self):
        defaultdict.__init__(self, SuffixTree)
        self.labels = set()

    def add_ngram(self, sequence, label=None):
        if label is None:
            label = sequence
        for i in range(len(sequence)):
            self.add(sequence[i:], label=label)


class KeyTransformingPrefixTree(PrefixTree):
    def __init__(self, transformer):
        defaultdict.__init__(self, lambda: KeyTransformingPrefixTree(transformer))
        self.transformer = transformer
        self.labels = set()

    def get(self, key):
        return self[self.transformer(key)]

    def add(self, sequence, label=None):
        layer = self
        t = self.transformer
        if label is None:
            label = sequence
        if label:
            layer.labels.add(label)
        for i in range(len(sequence)):
            layer = layer[t(sequence[i])]
            if label:
                layer.labels.add(label)
        return self

    def __contains__(self, sequence):
        layer = self
        j = 0
        t = self.transformer
        for i in sequence:
            j += 1
            if not dict.__contains__(layer, t(i)):
                break
            layer = layer[i]
        return j == len(sequence)

    def depth_in(self, sequence):
        layer = self
        count = 0
        t = self.transformer
        for i in sequence:
            if not dict.__contains__(layer, t(i)):
                break
            else:
                layer = layer[i]
            count += 1
        return count

    def subsequences_of(self, sequence):
        layer = self
        t = self.transformer
        for i in sequence:
            layer = layer[t(i)]
        return layer.labels

    __repr__ = dict.__repr__

    def __iter__(self):
        return iter(self.labels)

    @contextmanager
    def with_identity(self):
        t = self.transformer
        self.transformer = identity
        yield
        self.transformer = t


class KeyTransformingSuffixTree(KeyTransformingPrefixTree):
    def __init__(self, transformer):
        defaultdict.__init__(self, lambda: KeyTransformingSuffixTree(transformer))
        self.transformer = transformer
        self.labels = set()

    def add_ngram(self, sequence, label=None):
        if label is None:
            label = sequence
        for i in range(len(sequence)):
            self.add(sequence[i:], label=label)



cimport cython
from glypy.composition.ccomposition cimport CComposition
from glypy.composition import formula as _formula

from cpython.list cimport PyList_GetItem, PyList_GET_ITEM, PyList_Size, PyList_GET_SIZE
from cpython.sequence cimport PySequence_GetItem


cdef object formula = _formula

DEF SAFE_GET = 1

@cython.freelist(10000)
cdef class PeptideSequenceBase(object):

    cdef SequencePosition get(self, ssize_t i):
        IF SAFE_GET:
            return <SequencePosition>PyList_GetItem(self.sequence, i)
        ELSE:
            return <SequencePosition>PyList_GET_ITEM(self.sequence, i)

    cpdef _invalidate(self):
        pass


@cython.freelist(100)
cdef class TerminalGroup(object):
    def __init__(self, base_composition, modification=None):
        if not isinstance(base_composition, CComposition):
            base_composition = CComposition(base_composition)
        self.base_composition = base_composition
        self._modification = None
        if modification is not None:
            self.modification = modification
        self.mass = self._calculate_mass()

    def _calculate_mass(self):
        base_mass = self.base_composition.mass
        mod = self.modification
        if mod is not None:
            base_mass += mod.mass
        return base_mass

    def clone(self):
        return self.__class__(self.base_composition, self.modification)

    def __reduce__(self):
        return self.__class__, (self.base_composition, self.modification)


    cdef ModificationBase get_modification(self):
        return self._modification

    cdef void set_modification(self, ModificationBase value):
        cdef:
            double new_mass, old_mass
        if value is not None:
            new_mass = value.mass
        else:
            new_mass = 0
        if self._modification is not None:
            old_mass = self._modification.mass
        else:
            old_mass = 0
        self.mass += new_mass - old_mass
        self._modification = value

    @property
    def modification(self):
        return self.get_modification()

    @modification.setter
    def modification(self, ModificationBase value):
        self.set_modification(value)

    cpdef TerminalGroup modify(self, ModificationBase modification):
        return self.__class__(self.base_composition, modification)

    cdef CComposition get_composition(self):
        cdef ModificationBase modification = self.get_modification()

        if modification is None:
            return self.base_composition
        mod_comp = modification.composition
        return self.base_composition + mod_comp

    @property
    def composition(self):
        return self.get_composition()

    def __repr__(self):
        template = "{self.__class__.__name__}({self.base_composition}, {self.modification})"
        return template.format(self=self)

    def __str__(self):
        return self.serialize()

    def __eq__(self, other):
        if other is None:
            return False
        try:
            return (self.base_composition == other.base_composition) and (self.modification == other.modification)
        except AttributeError:
            if isinstance(other, basestring):
                from glycopeptidepy.structure.modification import Modification
                return self.composition == Modification(other).composition
            else:
                try:
                    return self.composition == other.composition
                except AttributeError:
                    return NotImplemented

    cdef bint equal_to(self, object other):
        cdef:
            TerminalGroup other_terminal
        if isinstance(other, TerminalGroup):
            other_terminal = <TerminalGroup>other
            return (self.base_composition == other_terminal.base_composition) and (
                self.modification == other_terminal.modification)
        else:
            if isinstance(other, basestring):
                from glycopeptidepy.structure.modification import Modification
                return self.composition == Modification(other).composition
            elif isinstance(other, ModificationBase):
                return self.composition == (<ModificationBase>other).composition
            else:
                try:
                    return self.composition == other.composition
                except AttributeError:
                    return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(formula(self.get_composition()))

    cpdef str serialize(self):
        if self.modification is not None:
            return str(self.modification)
        return formula(self.base_composition)


@cython.freelist(100)
cdef class AminoAcidResidueBase(object):
    '''
    A base type for classes describing amino acid residues
    '''

    def __copy__(self):
        return self.clone()

    @staticmethod
    cdef AminoAcidResidueBase _create(str name, str symbol, double mass, CComposition composition):
        cdef AminoAcidResidueBase inst = AminoAcidResidueBase.__new__(AminoAcidResidueBase)
        inst.name = name
        inst.symbol = symbol
        inst.mass = mass
        inst.composition = composition
        return inst

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if self is other:
            return True
        try:
            return self.name == other.name or self.symbol == other.symbol
        except AttributeError:
            return self.name == other or self.symbol == other

    cdef bint equal_to(self, AminoAcidResidueBase other):
        return self.name == other.name or self.symbol == other.symbol

    def __ne__(self, other):
        if self is other:
            return False
        try:
            return self.name != other.name and self.symbol != other.symbol
        except AttributeError:
            return self.name != other and self.symbol != other


@cython.freelist(100000)
cdef class ModificationBase(object):

    def __copy__(self):
        return self.clone()

    cpdef bint is_a(self, object category):
        '''Returns whether or not this :class:`ModificationBase` object belongs to
        the specified :class:`~.ModificationCategory`.

        Returns
        -------
        bool
        '''
        return False

    cpdef basestring serialize(self):
        '''A string representation for inclusion in sequences'''
        return self.name

    cpdef bint is_tracked_for(self, object category):
        """Determine if this :class:`ModificationBase` is tracked by a particular
        behavioral pattern associated with a :class:`~.ModificationCategory`.

        This relationship is distinct from :meth:`is_a` which merely observes that
        the semantic relationship holds, not that any actual behavior is available.

        Parameters
        ----------
        category : :class:`~.ModificationCategory`
            The category to check

        Returns
        -------
        bool
        """
        return False


@cython.final
@cython.freelist(10000000)
cdef class SequencePosition(object):

    def __init__(self, parts):
        self.amino_acid = PySequence_GetItem(parts, 0)
        self.modifications = PySequence_GetItem(parts, 1)

    def __iter__(self):
        yield self.amino_acid
        yield self.modifications

    def __getitem__(self, i):
        if i == 0:
            return self.amino_acid
        elif i == 1:
            return self.modifications
        else:
            raise IndexError(i)

    def __setitem__(self, i, value):
        if i == 0:
            self.amino_acid = value
        elif i == 1:
            self.modifications = value
        else:
            raise IndexError(i)

    def __eq__(self, other):
        cdef:
            SequencePosition other_t
            size_t n, i
        if isinstance(other, SequencePosition):
            other_t = other
            if not self.amino_acid.equal_to(other_t.amino_acid):
                return False
            n = self.get_modification_count()
            if n != other_t.get_modification_count():
                return False
            if self.modifications != other_t.modifications:
                return False
            return True
        else:
            return self.amino_acid == other.amino_acid and (self.modifications == other.modifications)

    def __ne__(self, other):
        return not (self == other)

    def __reduce__(self):
        return self.__class__, ((self.amino_acid, self.modifications),)

    def __len__(self):
        return 2

    def __repr__(self):
        if self.modifications is None:
            modifications = '-'
        else:
            modifications = "%r" % self.modifications
        return "[%r, %s]" % (self.amino_acid, modifications)

    @staticmethod
    cdef SequencePosition _create(AminoAcidResidueBase amino_acid, list modifications):
        cdef SequencePosition inst = SequencePosition.__new__(SequencePosition)
        inst.amino_acid = amino_acid
        inst.modifications = modifications
        return inst

    cdef inline double get_mass(self):
        cdef:
            double mass
            size_t i, n
            ModificationBase mod

        mass = 0
        mass += self.amino_acid.mass
        n = self.get_modification_count()
        for i in range(n):
            mod = self.get_modification(i)
            mass += mod.mass
        return mass

    cpdef bint has_modification(self, modification):
        cdef:
            size_t i, n
            ModificationBase mod

        n = self.get_modification_count()
        for i in range(n):
            mod = self.get_modification(i)
            if modification == mod:
                return True
        return False

    cpdef bint is_modified(self):
        return self.get_modification_count() > 0

    cpdef add_modification(self, modification):
        if self.modifications is None:
            self.modifications = [modification]
        else:
            self.modifications.append(modification)

    cpdef drop_modification(self, modification):
        cdef:
            size_t i, n,
            int index
            ModificationBase mod

        n = self.get_modification_count()
        index = -1
        for i in range(n):
            mod = <ModificationBase>PyList_GET_ITEM(self.modifications, i)
            if modification == mod:
                index = i
                break
        if index == -1:
            raise ValueError("Modification %s not found" % modification)
        return self.modifications.pop(index)

    @property
    def mass(self):
        return self.get_mass()

    cdef size_t get_modification_count(self):
        if self.modifications is None:
            return 0
        return PyList_GET_SIZE(self.modifications)

    cdef ModificationBase get_modification(self, size_t i):
        return <ModificationBase>PyList_GET_ITEM(self.modifications, i)


cdef extern from "compat.h":
    char* PyStr_AsString(str string)
    char* PyStr_AsUTF8AndSize(str string, Py_ssize_t*)
    str PyStr_FromString(char* string)
    long PyInt_AsLong(object i)
    object PyInt_FromLong(long i)
    str PyStr_Format(object format, object args)
    str PyStr_FromStringAndSize(char* string, Py_ssize_t size)
    cdef int IS_PY3


import itertools
from collections import defaultdict

from six import string_types as basestring

from glycopeptidepy.enzyme import Protease
from glycopeptidepy.structure.residue import UnknownAminoAcidException
from glycopeptidepy.structure.modification import SequenceLocation
from glycopeptidepy.structure.sequence import PeptideSequence, list_to_sequence
from glycopeptidepy.structure.parser import sequence_tokenizer_respect_sequons, sequence_tokenizer


def pair_rotate(sequence):
    """Invert each token pair.

    ABCD -> BADC

    Parameters
    ----------
    sequence : iterable

    Returns
    -------
    list
    """
    chunks = []
    gen = iter(sequence)
    while True:
        chunk = []
        try:
            chunk = [next(gen)]
            chunk.append(next(gen))
            chunks.append(chunk)
        except StopIteration:
            if len(chunk) > 0:
                chunks.append(chunk)
            break
    rev_seq = []
    for chunk in reversed(chunks):
        rev_seq.extend(chunk)
    return rev_seq


def edit_distance(s1, s2):
    if len(s1) > len(s2):
        s1, s2 = s2, s1

    distances = range(len(s1) + 1)
    for i2, c2 in enumerate(s2):
        distances_ = [i2 + 1]
        for i1, c1 in enumerate(s1):
            if c1 == c2:
                distances_.append(distances[i1])
            else:
                distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
        distances = distances_
    return distances[-1]


def longest_common_substring(seq1, seq2):
    n = len(seq1)
    m = len(seq2)
    solution_matrix = [[0 for i in range(m)] for j in range(n)]
    # track the length of the current longest common subsequence
    z = 0
    result = []
    for i in range(n):
        for j in range(m):
            if seq1[i] == seq2[j]:
                # match the start of one string
                if i == 0 or j == 0:
                    solution_matrix[i][j] = 1
                else:
                    solution_matrix[i][j] = solution_matrix[i - 1][j - 1] + 1
                # new longest substring
                if solution_matrix[i][j] > z:
                    z = solution_matrix[i][j]
                    result = list(seq1[i - z + 1:i + 1])
                # extending existing solution (may be invalid)
                elif solution_matrix[i][j] == z:
                    result.extend(list(seq1[i - z + 1:i + 1]))
            else:
                solution_matrix[i][j] = 0
    return result


def reverse_preserve_sequon(sequence, prefix_len=0, suffix_len=1, peptide_type=None, known_sites=None):
    if peptide_type is None:
        peptide_type = PeptideSequence
    if isinstance(sequence, PeptideSequence):
        sequence = str(sequence)
    original = peptide_type(sequence)
    sequence_tokens = sequence_tokenizer_respect_sequons(sequence, known_sites=known_sites)
    pref = sequence_tokens[:prefix_len]
    if suffix_len == 0:
        suf = []
        body = sequence_tokens[prefix_len:]
    else:
        suf = sequence_tokens[-suffix_len:]
        body = sequence_tokens[prefix_len:-suffix_len]
    body = body[::-1]
    rev_sequence = (list_to_sequence(list(pref) + list(body) + list(suf)))
    if str(list_to_sequence(sequence_tokens)) == str(rev_sequence):
        rot_body = pair_rotate(body)
        rev_sequence = (list_to_sequence(pref + list(rot_body) + suf))
    rev_sequence.n_term = original.n_term.clone()
    rev_sequence.c_term = original.c_term.clone()
    if original.glycan:
        rev_sequence.glycan = original.glycan.clone(propogate_composition_offset=False)
    return rev_sequence


def reverse_sequence(sequence, prefix_len=0, suffix_len=1, peptide_type=None, known_sites=None):
    if peptide_type is None:
        peptide_type = PeptideSequence
    if isinstance(sequence, PeptideSequence):
        sequence = str(sequence)
    original = peptide_type(sequence)
    sequence_tokens = sequence_tokenizer(sequence)[0]
    pref = sequence_tokens[:prefix_len]
    if suffix_len == 0:
        suf = []
        body = sequence_tokens[prefix_len:]
    else:
        suf = sequence_tokens[-suffix_len:]
        body = sequence_tokens[prefix_len:-suffix_len]
    body = body[::-1]
    rev_sequence = (list_to_sequence(list(pref) + list(body) + list(suf)))
    if str(list_to_sequence(sequence_tokens)) == str(rev_sequence):
        rot_body = pair_rotate(body)
        rev_sequence = (list_to_sequence(pref + list(rot_body) + suf))
    rev_sequence.n_term = original.n_term.clone()
    rev_sequence.c_term = original.c_term.clone()
    if original.glycan:
        rev_sequence.glycan = original.glycan.clone(propogate_composition_offset=False)
    return rev_sequence


class LimitedCrossproduct(object):
    def __init__(self, collections, max_depth=4):
        self.collections = collections
        self.max_depth = max_depth
        self.size = len(collections)

        self.precomputed = self.compose_iterative()
        self.iterator = iter(self.precomputed)

    def compose(self):
        i = 0
        current = []
        depth = 0
        for res in self.compose_inner(i, current, depth):
            yield res

    def compose_inner(self, layer_index, current, depth):
        if layer_index == self.size:
            yield current, depth
            return
        layer = self.collections[layer_index]
        for val in layer:
            new_depth = depth + (val is not None)
            if new_depth <= self.max_depth:
                for res in self.compose_inner(layer_index + 1, current + [val], new_depth):
                    yield res
            else:
                continue

    def compose_iterative(self):
        previous = [([], 0)]
        layer_index = 0
        for layer in self.collections:
            current = []
            for rec, depth in previous:
                for val in layer:
                    new_depth = depth + (val is not None)
                    if new_depth <= self.max_depth:
                        current.append((rec + [val], new_depth))
            previous = current
        return previous

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)




class ModificationSiteAssignmentCombinator(object):
    def __init__(self, variable_site_map, max_modifications=4):
        self.modification_to_site = variable_site_map
        self.site_to_modification = self.transpose_sites()
        self.max_modifications = max_modifications

    def transpose_sites(self):
        """Given a dictionary mapping between modification names and
        an iterable of valid sites, create a dictionary mapping between
        modification names and a list of valid sites plus the constant `None`

        Returns
        -------
        dict
        """
        sites = defaultdict(list)
        for mod, varsites in self.modification_to_site.items():
            for site in varsites:
                sites[site].append(mod)
        for site in list(sites):
            sites[site].append(None)
        return sites

    def _remove_empty_sites(self, site_mod_pairs):
        return [sm for sm in site_mod_pairs if sm[1] is not None]

    def assign(self):
        sites = list(self.site_to_modification.keys())
        choices = list(self.site_to_modification.values())
        for selected in LimitedCrossproduct(choices, max_depth=self.max_modifications):
            site_mod_pairs = zip(sites, selected)
            assignment = self._remove_empty_sites(site_mod_pairs)
            if assignment > self.max_modifications:
                continue
            yield assignment

    def __iter__(self):
        return self.assign()


class PeptidoformGenerator(object):
    def __init__(self, constant_modifications, variable_modifications, max_variable_modifications=None):
        if max_variable_modifications is None:
            max_variable_modifications = 4
        self.constant_modifications = list(constant_modifications)
        self.variable_modifications = list(variable_modifications)
        self.max_variable_modifications = max_variable_modifications

        (self.n_term_modifications,
         self.c_term_modifications,
         self.variable_modifications) = self.split_terminal_modifications(self.variable_modifications)

    @staticmethod
    def split_terminal_modifications(modifications):
        """Group modification rules into three classes, N-terminal,
        C-terminal, and Internal modifications.

        A modification rule can be assigned to multiple groups if it
        is valid at multiple sites.

        Parameters
        ----------
        modifications : Iterable of ModificationRule
            The modification rules

        Returns
        -------
        n_terminal: list
            list of N-terminal modification rules
        c_terminal: list
            list of C-terminal modification rules
        internal: list
            list of Internal modification rules
        """
        n_terminal = []
        c_terminal = []
        internal = []

        for mod in modifications:
            n_term = mod.n_term_targets
            if n_term and all([t.amino_acid_targets is None for t in n_term]):
                n_term_rule = mod.clone(n_term)
                mod = mod - n_term_rule
                n_terminal.append(n_term_rule)
            c_term = mod.c_term_targets
            if c_term and all([t.amino_acid_targets is None for t in c_term]):
                c_term_rule = mod.clone(c_term)
                mod = mod - c_term_rule
                c_terminal.append(c_term_rule)
            if (mod.targets):
                internal.append(mod)

        return n_terminal, c_terminal, internal

    def prepare_peptide(self, sequence):
        if not isinstance(sequence, PeptideSequence):
            return PeptideSequence(str(sequence))
        return sequence

    def terminal_modifications(self, sequence, protein_n_term=False, protein_c_term=False):
        n_term_modifications = [
            mod for mod in self.n_term_modifications if mod.find_valid_sites(
                sequence, protein_n_term=protein_n_term)]
        c_term_modifications = [
            mod for mod in self.c_term_modifications if mod.find_valid_sites(
                sequence, protein_c_term=protein_c_term)]
        # the case with unmodified termini
        n_term_modifications.append(None)
        c_term_modifications.append(None)
        return (n_term_modifications, c_term_modifications)

    def apply_fixed_modifications(self, sequence, protein_n_term=False, protein_c_term=False):
        has_fixed_n_term = False
        has_fixed_c_term = False

        for mod in self.constant_modifications:
            for site in mod.find_valid_sites(sequence, protein_n_term=protein_n_term,
                                             protein_c_term=protein_c_term):
                if site == SequenceLocation.n_term or site == SequenceLocation.protein_n_term:
                    has_fixed_n_term = True
                elif site == SequenceLocation.c_term or site == SequenceLocation.protein_c_term:
                    has_fixed_c_term = True
                sequence.add_modification(site, mod.name)
        return has_fixed_n_term, has_fixed_c_term

    def modification_sites(self, sequence):
        variable_sites = {
            mod.name: set(
                mod.find_valid_sites(sequence)) for mod in self.variable_modifications}
        modification_sites = ModificationSiteAssignmentCombinator(variable_sites)
        return modification_sites

    def apply_variable_modifications(self, sequence, assignments, n_term, c_term):
        n_variable = 0
        result = sequence.clone()
        if n_term is not None:
            result.n_term = n_term
            n_variable += 1
        if c_term is not None:
            result.c_term = c_term
            n_variable += 1
        for site, mod in assignments:
            if mod is not None:
                result.add_modification(site, mod)
                n_variable += 1
        return result, n_variable

    def generate_peptidoforms(self, sequence, protein_n_term=False, protein_c_term=False):
        try:
            sequence = self.prepare_peptide(sequence)
        except UnknownAminoAcidException:
            return
        (n_term_modifications,
         c_term_modifications) = self.terminal_modifications(
             sequence, protein_n_term=protein_n_term, protein_c_term=protein_c_term)

        (has_fixed_n_term,
         has_fixed_c_term) = self.apply_fixed_modifications(
             sequence, protein_n_term=protein_n_term, protein_c_term=protein_c_term)

        if has_fixed_n_term:
            n_term_modifications = [None]
        if has_fixed_c_term:
            c_term_modifications = [None]

        modification_sites = self.modification_sites(sequence)

        for n_term, c_term in itertools.product(n_term_modifications, c_term_modifications):
            for assignments in modification_sites:
                if len(assignments) > self.max_variable_modifications:
                    continue
                yield self.apply_variable_modifications(
                    sequence, assignments, n_term, c_term)

    def __call__(self, peptide, protein_n_term=False, protein_c_term=False):
        return self.generate_peptidoforms(
            peptide, protein_n_term=protein_n_term, protein_c_term=protein_c_term)

    @classmethod
    def peptidoforms(cls, sequence, constant_modifications, variable_modifications, max_variable_modifications=4):
        inst = cls(constant_modifications, variable_modifications,
                   max_variable_modifications)
        return inst(sequence)


class ProteinDigestor(object):

    def __init__(self, protease, constant_modifications=None, variable_modifications=None,
                 max_missed_cleavages=2, min_length=6, max_length=60, semispecific=False,
                 max_variable_modifications=None):
        if constant_modifications is None:
            constant_modifications = []
        if variable_modifications is None:
            variable_modifications = []
        self.protease = self._prepare_protease(protease)
        self.constant_modifications = constant_modifications
        self.variable_modifications = variable_modifications
        self.peptidoform_generator = PeptidoformGenerator(
            self.constant_modifications,
            self.variable_modifications,
            max_variable_modifications=max_variable_modifications)
        self.max_missed_cleavages = max_missed_cleavages
        self.min_length = min_length
        self.max_length = max_length
        self.semispecific = semispecific
        self.max_variable_modifications = max_variable_modifications

    def _prepare_protease(self, protease):
        if isinstance(protease, Protease):
            pass
        elif isinstance(protease, basestring):
            protease = Protease(protease)
        elif isinstance(protease, (list, tuple)):
            protease = Protease.combine(*protease)
        return protease

    def cleave(self, sequence):
        seqs = self.protease.cleave(
            sequence, self.max_missed_cleavages,
            min_length=self.min_length, max_length=self.max_length,
            semispecific=self.semispecific)
        for peptide, start, end, missed in seqs:
            if "X" in peptide:
                continue
            yield peptide, start, end, missed

    def _prepare_protein(self, protein):
        if isinstance(protein, PeptideSequence):
            return protein
        return PeptideSequence(str(protein))

    def digest(self, protein):
        sequence = self._prepare_protein(protein)
        for peptide, start, end, n_missed_cleavages in self.cleave(sequence):
            if end - start > self.max_length:
                continue
            for inst in self.modify_string(peptide):
                inst.count_missed_cleavages = n_missed_cleavages
                inst.start_position = start
                inst.end_position = end
                yield inst

    def modify_string(self, peptide):
        for modified_peptide, n_variable_modifications in self.peptidoform_generator(peptide):
            modified_peptide.count_variable_modifications = n_variable_modifications
            yield modified_peptide


try:
    _ModificationSiteAssignmentCombinator = ModificationSiteAssignmentCombinator
    has_c = True
    from glycopeptidepy._c.algorithm import ModificationSiteAssignmentCombinator, LimitedCrossproduct
except ImportError:
    has_c = False



from typing import List
from . import ResidueBase
from .composition import Composition
from glypy.utils.multimap import MultiMap
from six import add_metaclass


try:
    # check to see if the C base class exists so we
    # do not allocate additional slots
    _has_c = True
    from glycopeptidepy._c.structure import base
    del base
except ImportError:
    _has_c = False

symbol_to_residue = {
    'A': 'Ala',
    'R': 'Arg',
    'N': 'Asn',
    'D': 'Asp',
    'C': 'Cys',
    'E': 'Glu',
    'Q': 'Gln',
    'G': 'Gly',
    'H': 'His',
    'I': 'Ile',
    'L': 'Leu',
    'K': 'Lys',
    'M': 'Met',
    'F': 'Phe',
    'P': 'Pro',
    'S': 'Ser',
    'T': 'Thr',
    'W': 'Trp',
    'Y': 'Tyr',
    'V': 'Val',
    "U": "Sec",
    "O": "Pyl",
    "X": "Unk"
}


residue_to_symbol = {value: key for key, value in symbol_to_residue.items()}


symbol_to_long = {
    "A": "Alanine",
    "R": "Arginine",
    "N": "Asparagine",
    "D": "Aspartic Acid",
    "C": "Cysteine",
    "E": "Glutamic Acid",
    "Q": "Glutamine",
    "G": "Glycine",
    "H": "Histidine",
    "I": "Isoleucine",
    "L": "Leucine",
    "K": "Lysine",
    "M": "Methionine",
    "F": "Phenylalanine",
    "P": "Proline",
    "S": "Serine",
    "T": "Threonine",
    "W": "Tryptophan",
    "Y": "Tyrosine",
    "V": "Valine",
    "X": "Unknown",
    "U": "Selenocysteine",
    "O": "Pyrrolysine",
}


long_to_symbol = {value: key for key, value in symbol_to_long.items()}


residue_table = {
    'Ala': 'C3H5NO',
    'Arg': 'C6H12N4O1',
    'Asn': 'C4H6N2O2',
    'Asp': 'C4H5N1O3',
    'Cys': 'C3H5N1O1S1',
    'Glu': 'C5H7NO3',
    'Gln': 'C5H8N2O2',
    'Gly': 'C2H3N1O1',
    'His': 'C6H7N3O1',
    'Ile': 'C6H11N1O1',
    'Leu': 'C6H11N1O1',
    'Lys': 'C6H12N2O1',
    'Met': 'C5H9N1O1S1',
    'Phe': 'C9H9N1O1',
    'Pro': 'C5H7N1O1',
    'Ser': 'C3H5N1O2',
    'Thr': 'C4H7N1O2',
    'Trp': 'C11H10N2O1',
    'Tyr': 'C9H9N1O2',
    'Val': 'C5H9N1O1',
    "Sec": "C3H5NO1Se",
    "Pyl": "C12H19N3O2",
    "Unk": "",
}

residue_chemical_property_group = {
    'Ala': 'hydrophobic',
    'Arg': 'positive_side_chain',
    'Asn': 'polar_uncharged',
    'Asp': 'negative_side_chain',
    'Cys': 'special_case',
    'Glu': 'negative_side_chain',
    'Gln': 'polar_uncharged',
    'Gly': 'special_case',
    'His': 'positive_side_chain',
    'Ile': 'hydrophobic',
    'Leu': 'hydrophobic',
    'Lys': 'positive_side_chain',
    'Met': 'hydrophobic',
    'Phe': 'hydrophobic',
    'Pro': 'special_case',
    'Ser': 'polar_uncharged',
    'Thr': 'polar_uncharged',
    'Trp': 'hydrophobic',
    'Tyr': 'hydrophobic',
    'Val': 'hydrophobic',
}

residue_to_neutral_loss = MultiMap()
residue_to_neutral_loss.update({
    "Ser": [-Composition("H2O")],
    "Thr": [-Composition("H2O")],
    "Glu": [-Composition("H2O")],
    "Asp": [-Composition("H2O")],

    "Arg": [-Composition("NH3")],
    "Lys": [-Composition("NH3")],
    "Gln": [-Composition("NH3")],
    "Asn": [-Composition("NH3")],
})

residue_to_neutral_loss.update({
    "Arg": [-Composition("H2O")],
    "His": [-Composition("H2O")],
    "Lys": [-Composition("H2O")]
})


degeneracy_index = {
}


standard_residues = [
    'Ala',
    'Arg',
    'Asn',
    'Asp',
    'Cys',
    'Glu',
    'Gln',
    'Gly',
    'His',
    'Ile',
    'Leu',
    'Lys',
    'Met',
    'Phe',
    'Pro',
    'Ser',
    'Thr',
    'Trp',
    'Tyr',
    'Val',
]


class UnknownAminoAcidException(KeyError):
    pass


class MemoizedResidueMetaclass(type):
    '''
    A metaclass that memoizes Residues as they are constructed
    by overriding the class __call__ method. It will attempt to
    look up previously created residues by symbol, then by name.
    If a previous instance is not found, it will be created and
    saved.

    Attributes
    ----------
    _cache: dict

    '''

    def __call__(self, symbol=None, name=None, *args, **kwargs):
        try:
            cache = self._cache
        except AttributeError:
            cache = self._cache = dict()
        try:
            if symbol is not None:
                return cache[symbol]
            elif name is not None:
                return cache[name]
            else:
                raise Exception("Must provide a symbol or name parameter")
        except KeyError:
            if symbol is not None:
                inst = type.__call__(self, symbol=symbol, *args, **kwargs)
                cache[inst.symbol] = inst
                cache[inst.name] = inst
                return inst

            elif name is not None:
                inst = type.__call__(self, name=name, *args, **kwargs)
                cache[inst.symbol] = inst
                cache[inst.name] = inst
                return inst
            else:
                raise UnknownAminoAcidException(
                    "Cannot find a AminoAcidResidue for %r" % ((symbol, name),))

    def parse(self, symbol):
        try:
            cache = self._cache
            return cache[symbol]
        except (AttributeError, KeyError):
            return self(symbol)


def _AminoAcidResidue_reconstructor():
    return AminoAcidResidue.__new__(AminoAcidResidue)


@add_metaclass(MemoizedResidueMetaclass)
class AminoAcidResidue(ResidueBase):
    '''
    Represent a single Amino Acid residue which compose peptide sequences. The
    structure itself is intended to be immutable.

    Attributes
    ----------
    name: str
        The three letter abbreviation for the amino acid
    symbol: str
        The single letter abbreviation for the amino acid
    mass: float
        The neutral mass of the amino acid
    composition: :class:`glypy.Composition`
        The chemical composition of the amino acid
    chemical_class: str
        One of hydrophobic, polar_uncharged, positive_side_chain, negative_side_chain, or special_case,
        corresponding to the gross molecular properties of the amino acid
    is_degenerate: bool:
        Whether this amino acid may stand for other amino acids

    '''
    # If the C extension base class is available, do not declare redundant slot
    # descriptors.
    if _has_c:
        __slots__ = ["neutral_loss", ]
    else:
        __slots__ = ["name", "symbol", "mass", "composition", "_hash", "neutral_loss"]

    name: str
    symbol: str
    mass: float
    composition: Composition
    _hash: int
    neutral_loss: List

    def __init__(self, symbol=None, name=None):
        self.symbol = symbol
        self.name = name
        self.mass = 0.0
        try:
            if symbol is not None:
                self._by_symbol(symbol)
            elif name is not None:
                self._by_name(name)
        except KeyError:
            raise UnknownAminoAcidException(
                "No definition for Amino Acid %s" % (symbol if symbol is not None else name))
        self._hash = hash(self.name)
        self.neutral_loss = residue_to_neutral_loss[self.name]

    def _by_name(self, name: str):
        """Configure this instance by name information

        Parameters
        ----------
        name : str
            Amino Acid Name
        """
        self.composition = Composition(residue_table[name])
        self.name = name
        self.mass = self.composition.mass
        self.symbol = residue_to_symbol[name]

    def _by_symbol(self, symbol: str):
        """Configure this instance by symbol information,
        by going from symbol to name, and from name to data

        Parameters
        ----------
        symbol : str
            Amino Acid symbol
        """
        try:
            name = symbol_to_residue[symbol]
            self._by_name(name)
        except KeyError:
            self._by_name(symbol)

    def __repr__(self):
        return self.name

    if not _has_c:
        def __hash__(self):
            return self._hash

        def __eq__(self, other):
            if self is other:
                return True
            try:
                return self.name == other.name or self.symbol == other.symbol
            except AttributeError:
                return self.name == other or self.symbol == other

        def __ne__(self, other):
            if self is other:
                return False
            try:
                return self.name != other.name and self.symbol != other.symbol
            except AttributeError:
                return self.name != other and self.symbol != other

    def __reduce__(self):
        return _AminoAcidResidue_reconstructor, (), self.__getstate__()

    def __getstate__(self):
        return [self.name, self.symbol, self.mass, self.composition, self.neutral_loss]

    def __setstate__(self, state):
        self.name = state[0]
        self.symbol = state[1]
        self.mass = state[2]
        self.composition = state[3]
        if len(state) > 4:
            self.neutral_loss = state[4]
        self._hash = hash(self.name)

    @property
    def chemical_class(self):
        return residue_chemical_property_group[self.name]

    @property
    def is_degenerate(self):
        try:
            return degeneracy_index[self.name]
        except KeyError:
            return False

    @property
    def long_name(self):
        try:
            return symbol_to_long[self.symbol]
        except KeyError:
            return self.name


Residue = AminoAcidResidue


def register_residue(name, symbol, formula, chemical_class=None):
    assert symbol not in symbol_to_residue
    assert name not in residue_table
    residue_to_symbol[name] = symbol
    symbol_to_residue[symbol] = name
    residue_table[name] = formula
    residue_chemical_property_group[name] = chemical_class
    return AminoAcidResidue(symbol=symbol)


def register_degenerate(name, symbol, mappings):
    assert symbol not in symbol_to_residue
    assert name not in residue_table
    residue_to_symbol[name] = symbol
    symbol_to_residue[symbol] = name
    residue_table[name] = residue_table[mappings[0]]
    residue_chemical_property_group[
        name] = residue_chemical_property_group[mappings[0]]
    residue_to_neutral_loss[name] = residue_to_neutral_loss[mappings[0]]
    degeneracy_index[name] = frozenset(mappings)
    return AminoAcidResidue(symbol=symbol)


register_degenerate("Xle", "J", ["Leu", "Ile"])


def get_all_residues(omit_unknown=True, standard_only=True):
    '''Get a collection of all :class:`AminoAcidResidue` types defined.

    Parameters
    ----------
    omit_unknown: bool
        Whether to discard ``X`` or not.
    standard_only: bool
        Whether to include only the 20 standard amino acids or all
        defined amino acids.
    Returns
    -------
    set
    '''
    if standard_only:
        return set(map(AminoAcidResidue, standard_residues))
    residues = set(map(AminoAcidResidue, symbol_to_residue))
    if omit_unknown:
        residues.remove(AminoAcidResidue("X"))
    return residues


def get_all_sequencing_residues(omit_unknown=True, standard_only=True):
    residues = set(get_all_residues(omit_unknown, standard_only))
    residues.add(AminoAcidResidue("J"))

    for residue in list(residues):
        degenerate = residue.is_degenerate
        if degenerate:
            for degen in degenerate:
                residues.remove(degen)
    return residues


AminoAcidResidue.get_all_residues = staticmethod(get_all_residues)
AminoAcidResidue.get_all_sequencing_residues = staticmethod(get_all_sequencing_residues)




cdef class Configuration(object):
    def __init__(self):
        self.N_TERM_DEFAULT = "H"
        self.C_TERM_DEFAULT = "OH"
        self.FRAG_OFFSET = 1
        self.ALLOW_MODIFIED_ASPARAGINE = False
        self.PARTIAL_HEXNAC_LOSS = True
        self.EXCLUDE_B1 = True



from six import string_types as basestring

from glypy.utils import classproperty

from ..base import ModificationBase

from .source import ModificationTable
from .rule import ModificationRule


def _Modifcation_reconstructor():
    return Modification.__new__(Modification)


class ModificationInstanceBase(ModificationBase):
    def is_a(self, category):
        '''Returns whether or not this :class:`ModificationRule` object belongs to
        the specified :class:`~.ModificationCategory`.

        Returns
        -------
        bool
        '''
        return self.rule.is_a(category)

    def serialize(self):
        if self.rule.is_standard:
            rep = str(self.name)
        else:
            rep = self.rule.serialize()
        return rep

    def _init_from_rule(self, rule):
        self.name = rule.name
        self.mass = rule.mass
        self.rule = rule
        self._hash = self.rule._hash

        try:
            self.composition = rule.composition
        except AttributeError:
            self.composition = None

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        try:
            return other.name in self.rule.names
        except AttributeError:
            return other in self.rule.names

    def __ne__(self, other):
        return not self == other

    def is_tracked_for(self, category):
        """Determine if :attr:`rule` is tracked by a particular
        behavioral pattern associated with a :class:`~.ModificationCategory`.

        This relationship is distinct from :meth:`is_a` which merely observes that
        the semantic relationship holds, not that any actual behavior is available.

        Parameters
        ----------
        category : :class:`~.ModificationCategory`
            The category to check

        Returns
        -------
        bool
        """
        return self.rule.is_tracked_for(category)


try:
    _has_c = True
    from glycopeptidepy._c.structure.modification.modification import ModificationInstanceBase
except ImportError:
    _has_c = False


class Modification(ModificationInstanceBase):
    """Represents a peptide modification, with a fixed mass and optionally
    a composition.

    :class:`Modification` objects are usually bound to a :class:`~.ModificationRule`
    object which defines their properties. Though the essential properties are copied
    into the :class:`Modification` instance, the rule still holds additional metadata
    and behaviors.

    Modifications are equal to anything that their :attr:`rule` is equal to by name,
    and hash based upon their :attr:`name`.

    Attributes
    ----------
    name: str
        The preferred name of the modification
    mass: float
        The monosisotopic mass of the modification
    composition: :class:`~.Composition`
        The elemental composition of the modification. This
        attribute may be missing when the modification is not
        defined from a molecule but an arbitrary mass shift.
    rule: :class:`~.ModificationRule`
        The rule that defines the modification type this object represents
        an instance of.
    """

    # If the C extension base class is available, do not declare redundant slot
    # descriptors.
    if not _has_c:
        __slots__ = ["name", "mass", "rule", "composition", "_hash"]
    else:
        __slots__ = []

    _table = ModificationTable()

    @classmethod
    def register_new_rule(cls, rule):
        ModificationTable.register_new_rule(rule)

    def _resolve_name(self, rule_string):
        """Look up a :class:`~.ModificationRule` by name in :attr:`modification_table`.

        If the modification rule has not been registered with :attr:`modification_table`, and `rule_string`
        is not a full modification specification string, this may fail.

        Parameters
        ----------
        rule_string : :class:`str`
            The name or specification of a :class:`~.ModificationRule`
        """
        return self._table.resolve(rule_string)

    @classproperty
    def modification_table(self):
        """Access the class-wide :class:`~.ModificationTable` instance which all dynamic name
        lookups pass through.

        Returns
        -------
        :class:`~.ModificationTable`
        """
        return self._table

    @classmethod
    def from_rule(cls, rule):
        '''Create a new :class:`Modification` instance from a :class:`~.ModificationRule`
        instance directly.

        This method is marginally faster than the default initialization behavior which
        checks to see if the rule is a string.

        Parameters
        ----------
        rule: :class:`~.ModificationRule`
            The rule this modification will follow

        Returns
        -------
        :class:`Modification`
        '''
        inst = cls.__new__(cls)
        inst._init_from_rule(rule)
        return inst

    def __init__(self, rule):
        """Instantiate a :class:`Modification` instance from a template.

        Parameters
        ----------
        rule : :class:`str` or :class:`~.ModificationRule`
            Either the name of a modification rule as a string (to be resolved with
            :meth:`_resolve_name`) or an actual :class:`~.ModificationRule`, which will
            be used to populate all other attributes.
        """
        if(isinstance(rule, basestring)):
            rule = self._resolve_name(rule)

        self._init_from_rule(rule)

    @property
    def names(self):
        '''Returns the alternative names for this :class:`Modification`'s :attr:`rule`

        Returns
        -------
        frozenset
        '''
        return frozenset(self.rule.names)

    def __repr__(self):
        return self.serialize()

    def __str__(self):
        return self.serialize()

    def __reduce__(self):
        return _Modifcation_reconstructor, (), self.__getstate__()

    def __getstate__(self):
        return [self.name, self.mass, self.rule, self.composition]

    def __setstate__(self, state):
        self.name, self.mass, self.rule, self.composition = state
        self._hash = self.rule._hash

    def clone(self):
        return self.__class__(self.rule)

    def get_fragments(self, *args, **kwargs):
        return self.rule.get_fragments(*args, **kwargs)


# Late bind the ModificationRule's modification_tp field to avoid
# cyclic dependencies
ModificationRule.modification_tp = Modification



# cython: embedsignature=True

from libc.stdlib cimport malloc, free, realloc
from libc.string cimport strcpy, memcpy

cimport cython
from cpython cimport PyObject
from cpython.ref cimport Py_INCREF
from cpython.object cimport PyObject_Str
from cpython cimport PyErr_SetString
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_Append, PyList_GetItem, PyList_SetItem, PyList_New
from cpython.sequence cimport PySequence_Fast, PySequence_Fast_GET_ITEM, PySequence_Fast_GET_SIZE
from cpython.dict cimport PyDict_SetItem, PyDict_Keys, PyDict_Values, PyDict_Items, PyDict_Next, PyDict_GetItem
from cpython.int cimport PyInt_AsLong, PyInt_FromLong
from cpython.float cimport PyFloat_AsDouble
from cpython.tuple cimport PyTuple_GetItem

from glypy.composition.ccomposition cimport CComposition

from glycopeptidepy._c.structure.base cimport ModificationBase

from glycopeptidepy._c.compat cimport PyStr_AsUTF8AndSize, PyStr_FromStringAndSize

from glycopeptidepy._c.count_table cimport CountTable, CountTableIterator, count_table_items

from glycopeptidepy.structure.modification import (
    Modification, NGlycanCoreGlycosylation, OGlycanCoreGlycosylation,
    GlycosaminoglycanLinkerGlycosylation, ModificationCategory)

cdef ModificationBase _n_glycosylation = NGlycanCoreGlycosylation()
cdef ModificationBase _o_glycosylation = OGlycanCoreGlycosylation()
cdef ModificationBase _gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()
cdef ModificationBase _modification_hexnac = Modification("HexNAc").rule
cdef ModificationBase _modification_xylose = Modification("Xyl").rule


cdef struct string_cell:
    char* string
    size_t size

DEF ARRAY_SIZE = 2 ** 12
DEF DEFAULT_FRAGMENT_NAME_BUFFER_SIZE = 128

cdef string_cell[ARRAY_SIZE] str_ints
cdef dict str_int_overflow_intern_cache = {}

cdef void init_str_ints():
    cdef int i
    cdef Py_ssize_t z
    cdef str pa
    cdef char* a
    cdef char* atemp

    for i in range(ARRAY_SIZE):
        pa = str(i)
        atemp = PyStr_AsUTF8AndSize(pa, &z)
        a = <char*>malloc(sizeof(char) * z)
        strcpy(a, atemp)
        a[z] = "\0"
        str_ints[i].string = a
        str_ints[i].size = z


init_str_ints()


cdef int get_str_int(int i, string_cell* out):
    '''Get a `string_cell` for a given integer.

    For integers < 2 ** 12, go through the `str_ints` array
    fast path, otherwise use `str_int_overflow_intern_cache`.

    `str_int_overflow_intern_cache` keeps a mapping from PyInt to
    an immortalized PyStr whose internal buffer is used to serve
    the `char*` for `string_cell` instances larger than  2 ** 12 - 1.
    '''
    cdef:
        PyObject* ptemp
        object py_i
        str pa
        Py_ssize_t z
    if i < ARRAY_SIZE:
        out[0] = str_ints[i]
        return 0

    py_i = PyInt_FromLong(i)
    ptemp = PyDict_GetItem(str_int_overflow_intern_cache, py_i)
    if ptemp == NULL:
        pa = str(py_i)
        Py_INCREF(pa)
        PyDict_SetItem(str_int_overflow_intern_cache, py_i, pa)
    else:
        pa = <str>ptemp
    atemp = PyStr_AsUTF8AndSize(pa, &z)
    out.string = atemp
    out.size = z
    return 0


def _debug_get_str_int(int i):
    cdef:
        string_cell out
        int result
    result = get_str_int(i, &out)
    print(result)
    print(out.string)
    print(out.size)

def _get_string_cell(int i):
    cdef string_cell cell = str_ints[i]
    print(cell.size)
    print(cell.string)


@cython.freelist(100)
cdef class ChemicalShiftBase(object):

    cpdef ChemicalShiftBase clone(self):
        return self.__class__(self.name, self.composition.clone())

    def __str__(self):
        return self.name

    def __repr__(self):
        return "%s(name=%r)" % (self.__class__.__name__, self.name)

    def is_loss(self):
        return self.mass < 0


cdef class IonSeriesBase(object):

    def __eq__(self, other):
        cdef:
            IonSeriesBase other_tp
            object temp
        if not isinstance(self, IonSeriesBase):
            temp = self
            self = <IonSeriesBase>other
            other = temp
        if isinstance(other, IonSeriesBase):
            other_tp = <IonSeriesBase>other
            return self is other_tp or self.name == other_tp.name
        else:
            return self.name == other

    def __ne__(self, other):
        cdef:
            IonSeriesBase other_tp
            object temp
        if not isinstance(self, IonSeriesBase):
            temp = self
            self = <IonSeriesBase>other
            other = temp
        if isinstance(other, IonSeriesBase):
            other_tp = <IonSeriesBase>other
            return self.name != other_tp.name
        else:
            return self.name != other

    def __hash__(self):
        return self._hash


@cython.freelist(1000000)
cdef class FragmentBase(object):
    """Base class for all Fragment types. Defines basic
    name generation and neutral loss handling functions.

    Attributes
    ----------
    chemical_shift : ChemicalShift
        The ChemicalShift associated with this fragment, or None.
        If a ChemicalShift, its composition and mass are subtracted
        from this object's composition and mass attributes.
    name: str
        The human readable description of this fragment
    series: IonSeries
        The ion ladder this fragment is derived from

    """

    cpdef IonSeriesBase get_series(self):
        '''Return the :class:`IonSeries` this fragment belongs to.

        Returns
        -------
        :class:`IonSeries`
        '''
        raise NotImplementedError()

    def __hash__(self):
        if self._hash == -1:
            if self._name is None:
                self._update_hash_name()
            else:
                self._hash = hash(self._name)
        return self._hash

    def __eq__(self, other):
        try:
            return self.name == other.name and abs(self.mass - other.mass) < 1e-5
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other

    @property
    def series(self):
        '''The :class:`IonSeries` this fragment is drawn from.

        Returns
        -------
        :class:`IonSeries`
        '''
        return self.get_series()

    def clone(self):
        '''Create a copy of this object

        Returns
        -------
        :class:`FragmentBase`
        '''
        raise NotImplementedError()

    cpdef ChemicalShiftBase get_chemical_shift(self):
        '''Returns the chemical shift associated with the fragment.

        Returns
        -------
        :class:`ChemicalShift`
        '''
        return self._chemical_shift

    cpdef set_chemical_shift(self, ChemicalShiftBase chemical_shift):
        '''Sets the chemical shift associated with the fragment, updating
        the :attr:`mass` and :attr:`chemical_shift`.
        '''
        if self._chemical_shift is None and chemical_shift is None:
            return
        if self._chemical_shift is not None:
            self.mass -= self._chemical_shift.mass
        self._chemical_shift = chemical_shift
        if chemical_shift is not None:
            self.mass += chemical_shift.mass
        self._name = self.get_fragment_name()
        self._hash = hash(self._name)

    cdef void _update_hash_name(self):
        self._name = self.get_fragment_name()
        self._hash = hash(self._name)

    chemical_shift = property(get_chemical_shift, set_chemical_shift)

    cpdef str get_fragment_name(self):
        parts = [self._name]
        chemical_shift = self.get_chemical_shift()
        if chemical_shift is not None:
            parts.append(chemical_shift.name)
        return ''.join(parts)

    cpdef str base_name(self):
        """Simply return string like b2, y3 with no modification information."""
        return self._name

    cdef str get_name(self):
        if self._name is None:
            self._name = self.get_fragment_name()
        return self._name

    @property
    def name(self):
        return self.get_name()

    @name.setter
    def name(self, name):
        self._name = name


cdef object ModificationCategory_glycosylation = ModificationCategory.glycosylation


cdef class PeptideFragment(FragmentBase):
    '''Represents a peptide backbone fragment, such as a peptide b or y ion.

    Attributes
    ----------
    series: :class:`IonSeries`
        The ion series this fragment is from
    bare_mass: float
        The unmodified mass of this fragment
    mass: float
        The modified, correct mass of this fragment
    position: int
        The peptide backbone position index this fragment came from, relative to the
        starting terminal
    modification_dict: :class:`~.ModificationIndex`
        A count of the modifications on this fragment
    flanking_amino_acids: list
        The adjacent amino acids of the amide bond that broke to form this fragment
    glycosylation: dict
        A position to glycan mapping for glycan structures
    chemical_shift: :class:`ChemicalShift`
        An associated chemical shift
    composition: :class:`~.Composition`
        The elemental composition of this fragment
    '''
    @staticmethod
    cdef PeptideFragment _create(IonSeriesBase kind, int position, CountTable modification_dict, double mass,
                                 list flanking_amino_acids=None, dict glycosylation=None,
                                 ChemicalShiftBase chemical_shift=None, CComposition composition=None,
                                 double* delta_mass=NULL):
        cdef PeptideFragment self = PeptideFragment.__new__(PeptideFragment)
        self.kind = kind
        self.position = position

        self.bare_mass = mass
        self.mass = mass
        self.modification_dict = modification_dict
        self.composition = composition
        self._chemical_shift = None

        self.flanking_amino_acids = flanking_amino_acids
        self.glycosylation = glycosylation
        if chemical_shift is not None:
            self.set_chemical_shift(chemical_shift)

        if delta_mass == NULL:
            self._update_mass_with_modifications()
        else:
            self.mass += delta_mass[0]
        # if delta_mass != NULL:
        #     assert abs((self.mass - self.bare_mass) - delta_mass[0]) < 1e-3, ((self.mass - self.bare_mass), delta_mass[0])
        self._name = None
        self._hash = -1
        return self

    def __init__(self, kind, position, modification_dict, mass, flanking_amino_acids=None,
                 glycosylation=None, chemical_shift=None, composition=None):
        self.kind = kind

        # The mass value is the bare backbone's mass
        self.bare_mass = PyFloat_AsDouble(mass)
        self.modification_dict = modification_dict
        self.mass = self.bare_mass
        self.composition = composition
        self._chemical_shift = None

        self.flanking_amino_acids = flanking_amino_acids
        self.position = position

        self.glycosylation = glycosylation
        self.set_chemical_shift(chemical_shift)
        self._update_mass_with_modifications()

        self._name = None
        self._hash = -1

    cdef void _update_mass_with_modifications(self):
        cdef:
            ModificationBase mod
            CountTable modification_dict
            CountTableIterator modification_iterator
            PyObject *pkey
            long count
            Py_ssize_t ppos = 0

        modification_dict = self.modification_dict
        modification_iterator = CountTableIterator._create(modification_dict)

        while modification_iterator.has_more():
            ppos = modification_iterator.get_next_value(&pkey, &count)
            if ppos != 0:
                break
            mod = <ModificationBase>pkey
            self.mass += mod.mass * count

    cpdef IonSeriesBase get_series(self):
        return self.kind

    cpdef clone(self):
        return PeptideFragment._create(
            self.series, self.position, self.modification_dict.copy(),
            self.bare_mass, list(self.flanking_amino_acids),
            self.glycosylation.copy() if self.glycosylation is not None else None,
            self._chemical_shift.clone() if self._chemical_shift is not None else None,
            self.composition.clone() if self.composition is not None else None)

    def total_composition(self):
        if self.composition is None:
            raise ValueError("Composition data is missing")
        composition = self.composition.clone()
        chemical_shift = self.chemical_shift
        if chemical_shift is not None:
            composition += chemical_shift.composition
        return composition

    def __reduce__(self):
        return self.__class__, (
            self.kind, self.position, self.modification_dict, self.bare_mass,
            self.flanking_amino_acids, self.glycosylation,
            self.chemical_shift, self.composition)

    cpdef str base_name(self):
        """Simply return string like b2, y3 with no modification information."""
        fragment_name = []
        fragment_name.append(self.get_series().name)
        fragment_name.append(str(self.position))
        return ''.join(fragment_name)

    cpdef str get_fragment_name(self):
        """Build a complete fragment name listing any glycosylation modifications
        """
        cdef:
            ModificationBase mod_rule
            long count
            ChemicalShiftBase chemical_shift
            CountTable modification_dict
            CountTableIterator modification_iterator
            str name
            PyObject *pkey
            PyObject *pvalue
            Py_ssize_t index, size_ref, buffer_size
            Py_ssize_t ppos = 0

            string_cell int_conv

            char* tmp_buffer
            char[DEFAULT_FRAGMENT_NAME_BUFFER_SIZE] default_name_buffer
            char* name_buffer
            char* oversized_buffer
            bint needs_free

        name_buffer = <char*>default_name_buffer
        buffer_size = DEFAULT_FRAGMENT_NAME_BUFFER_SIZE
        needs_free = False
        index = 0
        tmp_buffer = PyStr_AsUTF8AndSize(self.get_series().name, &size_ref)
        memcpy(&name_buffer[index], tmp_buffer, size_ref)
        index += size_ref

        get_str_int(self.position, &int_conv)
        memcpy(&name_buffer[index], int_conv.string, int_conv.size)
        index += int_conv.size

        modification_dict = self.modification_dict
        modification_iterator = CountTableIterator._create(modification_dict)
        while modification_iterator.has_more():
            ppos = modification_iterator.get_next_value(&pkey, &count)
            if ppos != 0:
                break
            mod_rule = <ModificationBase>pkey
            if buffer_size - index < 30:
                if needs_free:
                    name_buffer = <char*>realloc(name_buffer, sizeof(char) * buffer_size * 2)
                    buffer_size *= 2
                else:
                    oversized_buffer = <char*>malloc(sizeof(char) * buffer_size * 2)
                    memcpy(oversized_buffer, name_buffer, index)
                    name_buffer = oversized_buffer
                    needs_free = True
                    buffer_size *= 2
            if mod_rule.is_a(ModificationCategory_glycosylation):
                if count > 1:
                    name_buffer[index] = "+"
                    index += 1
                    int_conv = str_ints[count]
                    memcpy(&name_buffer[index], int_conv.string, int_conv.size)
                    index += int_conv.size
                    tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
                    if size_ref + index > buffer_size:
                        if needs_free:
                            name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
                            buffer_size = (buffer_size + size_ref) * 2
                        else:
                            oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)
                            memcpy(oversized_buffer, name_buffer, index)
                            name_buffer = oversized_buffer
                            needs_free = True
                            buffer_size = (buffer_size + size_ref) * 2
                    memcpy(&name_buffer[index], tmp_buffer, size_ref)
                    index += size_ref
                elif count == 1:
                    name_buffer[index] = "+"
                    index += 1
                    tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
                    if size_ref + index > buffer_size:
                        if needs_free:
                            name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
                            buffer_size = (buffer_size + size_ref) * 2
                        else:
                            oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)
                            memcpy(oversized_buffer, name_buffer, index)
                            name_buffer = oversized_buffer
                            needs_free = True
                            buffer_size = (buffer_size + size_ref) * 2
                    memcpy(&name_buffer[index], tmp_buffer, size_ref)
                    index += size_ref

        chemical_shift = self.get_chemical_shift()
        if chemical_shift is not None:
            tmp_buffer = PyStr_AsUTF8AndSize(chemical_shift.name, &size_ref)
            if size_ref + index > buffer_size:
                if needs_free:
                    name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref))
                    buffer_size = (buffer_size + size_ref)
                else:
                    oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref))
                    memcpy(oversized_buffer, name_buffer, index)
                    name_buffer = oversized_buffer
                    needs_free = True
                    buffer_size = (buffer_size + size_ref)
            memcpy(&name_buffer[index], tmp_buffer, size_ref)
            index += size_ref
        name = PyStr_FromStringAndSize(&name_buffer[0], index)
        if needs_free:
            free(name_buffer)
        return name

    cdef bint _is_glycosylated(self):
        cdef:
            PyObject *pkey
            long value
            Py_ssize_t ppos = 0
            ModificationBase mod
            CountTable modification_dict
            CountTableIterator modification_iterator
        if self.glycosylation:
            return True
        else:
            modification_dict = self.modification_dict
            modification_iterator = CountTableIterator._create(modification_dict)
            while modification_iterator.has_more():
                ppos = modification_iterator.get_next_value(&pkey, &value)
                if ppos != 0:
                    break
                mod = <ModificationBase>pkey
                if mod.is_a(ModificationCategory_glycosylation):
                    return True
        return False

    @property
    def is_glycosylated(self):
        return self._is_glycosylated()

    @property
    def glycosylation_size(self):
        if self.glycosylation is not None:
            raise NotImplementedError()
        size = 0
        size += self.modification_dict.get(_modification_hexnac, 0)
        size += self.modification_dict.get(_modification_xylose, 0)
        return size

    cdef long get_glycosylation_size(self) except -1:
        cdef:
            long size
        if self.glycosylation is not None:
            PyErr_SetString(NotImplementedError, "Method not implemented for glycan collections yet")
            return -1
        size = 0
        size += self.modification_dict.getitem(_modification_hexnac)
        size += self.modification_dict.getitem(_modification_xylose)
        return size

    def __repr__(self):
        template = ("PeptideFragment({self.series}, {self.position}, "
                    "{self.mass}, {modification_dict}, {self.flanking_amino_acids},"
                    " {self.chemical_shift})")
        return template.format(self=self, modification_dict=dict(self.modification_dict))


cdef class SimpleFragment(FragmentBase):

    def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False):
        self._name = name
        self._hash = hash(self._name)
        self._chemical_shift = None

        self.mass = mass
        self.kind = kind
        self.composition = composition

        self.set_chemical_shift(chemical_shift)
        self.is_glycosylated = is_glycosylated

    cpdef clone(self):
        corrected_mass = self.mass
        if self._chemical_shift is not None:
            corrected_mass -= self.chemical_shift.mass
        return self.__class__(self.name, corrected_mass, self.kind,
                              self.composition,
                              self._chemical_shift.clone() if self._chemical_shift is not None else None,
                              self.is_glycosylated)

    def __reduce__(self):
        corrected_mass = self.mass
        if self._chemical_shift is not None:
            corrected_mass -= self.chemical_shift.mass
        return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
                                self.chemical_shift, self.is_glycosylated)

    def __repr__(self):
        return ("{self.__class__.__name__}(name={self.name}, "
                "mass={self.mass:.04f}, series={self.kind})").format(self=self)

    cpdef IonSeriesBase get_series(self):
        return self.kind


@cython.final
cdef class _NameTree(object):
    @staticmethod
    cdef _NameTree _create():
        cdef _NameTree self = _NameTree.__new__(_NameTree)
        self.name = None
        self.value = None
        self.children = dict()
        return self

    def __init__(self, name=None, children=None):
        if children is None:
            children = dict()
        self.name = name
        self.value = None
        self.children = children

    def __getitem__(self, key):
        return self.get(key)

    cpdef _NameTree get(self, key):
        cdef:
            PyObject* presult
            _NameTree result
        presult = PyDict_GetItem(self.children, key)
        if presult == NULL:
            result = _NameTree._create()
            PyDict_SetItem(self.children, key, result)
            return result
        else:
            result = <_NameTree>presult
            return result

    cpdef traverse(self, list_or_iterable parts):
        cdef:
            _NameTree node
            size_t i, n
            object parts_fast

        node = self
        if list_or_iterable is list:
            n = PyList_GET_SIZE(parts)
            for i in range(n):
                kv = <object>PyList_GET_ITEM(parts, i)
                node = node.get(kv)
        else:
            parts_fast = PySequence_Fast(parts, "Expected a sequence")
            n = PySequence_Fast_GET_SIZE(parts_fast)
            for i in range(n):
                kv = <object>PySequence_Fast_GET_ITEM(parts_fast, i)
                node = node.get(kv)
        return node


cdef _NameTree stub_fragment_name_cache = _NameTree()


cdef object pair_mass(tuple x):
    tmp = <object>PyTuple_GetItem(x, 0)
    return tmp.mass()


cdef basestring build_name_from_composition(mapping_types glycan_composition):
        cdef:
            basestring name, extended_key
            _NameTree root, node
            list parts
        name = 'peptide'
        root = stub_fragment_name_cache
        if mapping_types is dict:
            parts = PyDict_Items(glycan_composition)
        elif mapping_types is CountTable:
            parts = count_table_items(glycan_composition.table)
        elif mapping_types is object:
            parts = list(glycan_composition.items())
        # traverse the tree to find the node referring to
        # this combination of monosaccharides and counts
        node = root.traverse(parts)
        extended_key = node.name
        if extended_key is None:
            parts.sort(key=pair_mass)
            node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
        if extended_key:
            name = "%s+%s" % (name, extended_key)
        return name


cdef class StubFragment(FragmentBase):

    _name_cache = stub_fragment_name_cache

    @staticmethod
    cdef StubFragment _create(str name, double mass, IonSeriesBase kind, CComposition composition,
                              ChemicalShiftBase chemical_shift, bint is_glycosylated, object glycosylation,
                              bint is_extended, int glycosylation_size=-1):
        cdef:
            StubFragment self

        self = StubFragment.__new__(StubFragment)
        self._name = name
        self._hash = hash(self._name)
        self._chemical_shift = None
        self._glycosylation_size = glycosylation_size

        self.mass = mass
        self.kind = kind
        self.composition = composition
        if chemical_shift is not None:
            StubFragment.set_chemical_shift(self, chemical_shift)

        self.is_glycosylated = is_glycosylated
        self.glycosylation = glycosylation
        self.is_extended = is_extended

        return self

    def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False,
                 glycosylation=None, is_extended=False):
        self._name = name
        self._hash = hash(self._name)
        self._chemical_shift = None

        self.mass = mass
        self.kind = kind
        self.composition = composition

        self.set_chemical_shift(chemical_shift)
        self.is_glycosylated = is_glycosylated

        self.glycosylation = glycosylation
        self.is_extended = is_extended
        self._glycosylation_size = -1

    cpdef clone(self):
        cdef StubFragment dup = StubFragment._create(
            self.name, self.mass, self.kind, self.composition, self._chemical_shift,
            self.is_glycosylated, self.glycosylation, self.is_extended,
            self._glycosylation_size)
        return dup

    cdef int get_glycosylation_size(self):
        if self._glycosylation_size == -1:
            self._glycosylation_size = sum(self.glycosylation.values())
        return self._glycosylation_size

    @property
    def glycosylation_size(self):
        return self.get_glycosylation_size()

    cpdef str base_name(self):
        """Simply return the base fragment's name, omitting any chemical modification,
        e.g. peptide+HexNAc1"""

        # If the chemical shift is :const:`None`, there was no change to :attr:`_name`,
        # so just return it, otherwise we need to recompute the name from scratch using
        # :attr:`glycosylation` and :meth:`build_name_from_composition`
        if self._chemical_shift is None:
            return self._name
        else:
            return self.build_name_from_composition(self.glycosylation)

    @classmethod
    def build_name_from_composition(cls, glycan_composition):
        return build_name_from_composition(glycan_composition)

    def __reduce__(self):
        corrected_mass = self.mass
        if self._chemical_shift is not None:
            corrected_mass -= self.chemical_shift.mass
        return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
                                self.chemical_shift, self.is_glycosylated, self.glycosylation,
                                self.is_extended)

    def __repr__(self):
        return ("{self.__class__.__name__}(name={self.name}, "
                "mass={self.mass:.04f}, series={self.kind})").format(self=self)

    cpdef IonSeriesBase get_series(self):
        return self.kind



from cpython cimport Py_INCREF
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_Append, PyList_GetItem, PyList_SetItem, PyList_New
from cpython.dict cimport PyDict_SetItem, PyDict_Keys, PyDict_Values
from cpython.int cimport PyInt_AsLong
from cpython.tuple cimport PyTuple_GetItem
from cpython.mapping cimport PyMapping_Keys, PyMapping_Values

from glycopeptidepy._c.count_table cimport CountTable

cdef object product

from itertools import product

cpdef list descending_combination_counter(mapping_t counter):
    cdef:
        list keys, values, results
        object interval, v
        list count_ranges, combos
        int i, j, n, k
        CountTable result
        tuple combo
    if mapping_t is dict:
        keys = list(PyDict_Keys(counter))
        values = list(PyDict_Values(counter))
    elif mapping_t is CountTable:
        keys = counter.keys()
        values = counter.values()
    else:
        keys = list(PyMapping_Keys(counter))
        values = list(PyMapping_Values(counter))
    n = PyList_GET_SIZE(keys)
    count_ranges = PyList_New(n)
    for i in range(n):
        v = (<object>PyList_GET_ITEM(values, i))
        interval = range(v + 1)
        Py_INCREF(interval)
        PyList_SetItem(count_ranges, i, interval)

    combos = list(product(*count_ranges))
    k = PyList_GET_SIZE(combos)
    results = PyList_New(k)
    for i in range(k):
        result = CountTable._create()
        combo = <tuple>PyList_GET_ITEM(combos, i)
        for j in range(n):
            result.setitem(
                <object>PyList_GET_ITEM(keys, j),
                PyInt_AsLong(<object>PyTuple_GetItem(combo, j)))
        Py_INCREF(result)
        PyList_SetItem(results, i, result)
    return results



from glycopeptidepy._c.structure.modification.rule cimport ModificationRuleBase

cdef class ModificationTableBase(object):
    cdef:
        public dict store
        public dict _instance_custom_rules

    cpdef ModificationRuleBase resolve(self, basestring key)
    cpdef get_modification(self, name)

    cdef ModificationRuleBase _find_modification_rule(self, basestring key)
    cdef ModificationRuleBase _find_modification_rule_fuzzy(self, basestring key)


'''Represent queryable databases of peptide modifications from diverse
sources.
'''
import csv
import json
import bisect

from io import StringIO
from copy import deepcopy
from typing import List

from pkg_resources import resource_stream

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from ..composition import Composition

from .target import title_cleaner, extract_targets_from_string
from .rule import ModificationRule
from .utils import ModificationStringParseError
from .glycosylation import (
    hexnac_modification, xylose_modification,
    GlycosaminoglycanLinkerGlycosylation,
    OGlcNAcylation,
    OGlycanCoreGlycosylation,
    NGlycanCoreGlycosylation,)


def load_from_csv(stream):
    """Load a sequence of :class:`~.ModificationRule` objects from
    a CSV stream

    Parameters
    ----------
    stream : file-like
        A file-like object over a CSV

    Returns
    -------
    list of :class:`~.ModificationRule`
    """
    modification_definitions = []
    parser = csv.DictReader(stream)
    modification_definitions = [ModificationRule(**rec) for rec in parser]
    return modification_definitions


def load_from_json(stream):
    """Load a sequence of :class:`~.ModificationRule` objects from
    a JSON stream.

    :class:`~.ModificationRule` objects are constructed using the
    :meth:`~.ModificationRule.from_unimod` constructor instead of
    the default flat constructor.

    Parameters
    ----------
    stream : file-like
        A file-like object over a JSON

    Returns
    -------
    list of :class:`~.ModificationRule`
    """
    modification_definitions = []
    modification_definitions = list(map(ModificationRule.from_unimod, json.load(stream)))
    return modification_definitions


class ModificationSource(object):
    """An abstract collection of :class:`~.ModificationRule` objects.

    This class primarily deals with loading rules from file sources and
    managing global states. It should not be used instantiated directly,
    instead, use :class:`ModificationTable`.
    """

    _table_definition_file = staticmethod(lambda: StringIO(
        resource_stream(__name__, "data/ProteinProspectorModifications-for_gly2.csv").read(
        ).decode("utf-8")))
    _unimod_definitions = staticmethod(lambda: StringIO(
        resource_stream(__name__, "data/unimod.json").read().decode('utf-8')))

    use_protein_prospector = True

    @classmethod
    def _definitions_from_stream(cls, stream, format="csv"):
        if format == "csv":
            definitions = load_from_csv(stream)
        elif format == "json":
            definitions = load_from_json(stream)
        return definitions

    @classmethod
    def _definitions_from_stream_default(cls):
        defs = []
        defs += load_from_json(cls._unimod_definitions())
        if cls.use_protein_prospector:
            defs += load_from_csv(cls._table_definition_file())
        return defs

    @classmethod
    def load_from_file(cls, stream=None, format="csv"):
        '''Load the rules definitions from a CSV or JSON file and instantiate a
        :class:`ModificationSource` from it'''
        if stream is None:
            return cls.load_from_file_default()
        if format == "csv":
            definitions = load_from_csv(stream)
        elif format == "json":
            definitions = load_from_json(stream)
        return cls(definitions)

    @classmethod
    def load_from_file_default(cls):
        '''Load the rules definitions from the default package files'''
        defs = []
        defs += load_from_json(cls._unimod_definitions())
        if cls.use_protein_prospector:
            defs += load_from_csv(cls._table_definition_file())
        return cls(defs)

    bootstrapped = None

    @classmethod
    def bootstrap(cls, reuse=True):
        '''Load the bootstrapping file's definitions and instantiate a ModificationSource from it'''
        if cls.bootstrapped is not None and reuse:
            return cls.bootstrapped
        instance = cls.load_from_file()
        cls.bootstrapped = instance
        return instance

    def resolve(self, key):
        """Locate a :class:`~.Modification` instance from
        `key`.

        May call :meth:`~.ModificationRule.resolve`.

        Parameters
        ----------
        key : str
            The name of the :class:`~.ModificationRule`

        Returns
        -------
        :class:`~.ModificationRule`
        """
        raise NotImplementedError()

    def __getitem__(self, key):
        raise NotImplementedError()

    def get_modification(self, name):
        """Create a :class:`~.Modification` from the specified name

        Parameters
        ----------
        name : :class:`str`
            The name of the :class:`~.ModificationRule` to base the :class:`~.Modification`.

        Returns
        -------
        :class:`~.Modification`
        """
        return self.resolve(name)()

    def __call__(self, key):
        """Create a :class:`~.Modification` from the specified name

        Parameters
        ----------
        name : :class:`str`
            The name of the :class:`~.ModificationRule` to base the :class:`~.Modification`.

        Returns
        -------
        :class:`~.Modification`

        See Also
        --------
        :meth:`get_modification`
        """
        return self.get_modification(key)


try:
    from glycopeptidepy._c.structure.modification.source import ModificationTableBase
except ImportError as err:

    class ModificationTableBase(object):
        '''A mixin class to provide implementation for :class:`ModificationSource`
        methods for :class:`ModificationTable`
        '''

        def __getitem__(self, key):
            try:
                # First try the key in the custom mapping
                if key in self._instance_custom_rules:
                    return self._instance_custom_rules[key]
                else:
                    # Next try the normal store
                    return self.store[key]
            except KeyError:
                try:
                    # Clean the key in case it is decorated with extra target information
                    name = title_cleaner.search(key).groupdict()["name"]
                except (AttributeError):
                    raise ModificationStringParseError(
                        "Could not parse {0}".format(key))
                try:
                    return self.store[name]
                except KeyError:
                    try:
                        return self._instance_custom_rules[name]
                    except KeyError:
                        raise ModificationStringParseError(
                            "Could not resolve {0}".format(name))

        def resolve(self, key):
            try:
                rule = self[key]
            except ModificationStringParseError:
                rule = ModificationRule.resolve(key)
            return rule


class ModificationTable(ModificationTableBase, ModificationSource):
    '''A :class:`~.Mapping` connecting modification names to :class:`~.ModificationRule`
    instances.

    '''
    other_modifications = {
        "HexNAc": hexnac_modification,
        "Xyl": xylose_modification,
        "N-Glycosylation": NGlycanCoreGlycosylation(),
        "O-Glycosylation": OGlycanCoreGlycosylation(),
        "GAG-Linker": GlycosaminoglycanLinkerGlycosylation(),
        "O-GlcNAc": OGlcNAcylation(),
        "H": ModificationRule(
            "N-term", "H", "H",
            Composition("H").mass, composition=Composition("H")),
        "OH": ModificationRule(
            "C-term", "OH", "OH",
            Composition("OH").mass, composition=Composition("OH")),
    }

    _custom_rules = {}

    @classmethod
    def register_custom_rule(cls, rule):
        cls._custom_rules[rule.name] = rule

    register_new_rule = register_custom_rule

    @classmethod
    def remove_custom_rule(cls, rule):
        try:
            cls._custom_rules.pop(rule.name)
        except AttributeError:
            cls._custom_rules.pop(rule)

    def __init__(self, rules=None):
        if rules is None:
            rules = self._definitions_from_stream_default()
        self.store = dict()
        # For
        self._instance_custom_rules = self.__class__._custom_rules

        for rule in rules:
            if isinstance(rule, Mapping):
                rule = ModificationRule(**rule)
            if rule.name in self.other_modifications:
                continue
            self.add(rule)

        self._include_other_rules()

    def __len__(self):
        return len(self.store) + len(self._instance_custom_rules) + len(self.other_modifications)

    def __iter__(self):
        for key in iter(self.store):
            yield key
        for key in iter(self.other_modifications):
            yield key
        for key in iter(self._instance_custom_rules):
            yield key

    def _include_other_rules(self):
        for _name, rule in self.other_modifications.items():
            self.add(rule)

    def register_alias(self, name, alias):
        """Add a new name to an existing rule, and make that name
        resolvable.

        Parameters
        ----------
        name : :class:`str`
            The name of an existing :class:`~.ModificationRule`
        alias : :class:`str`
            The new name to add to the rule denoted by `name`.
        """
        rule = self[name]
        rule.aliases.add(alias)
        self.store[alias] = rule

    def rules(self):
        """The set of all :class:`~.ModificationRule` instances in this object

        Returns
        -------
        set
        """
        return set(self.store.values()) | set(self._custom_rules.values())

    def add(self, rule):
        """Add a new :class:`~.ModificationRule` to this object.

        If `rule` shares names with other rules and their masses match,
        they will be merged.

        Parameters
        ----------
        rule : :class:`~.ModificationRule` or :class:`Mapping`
            The rule to add. If is a :class:`Mapping`, it will be coerced
            into a :class:`~.ModificationRule`
        """
        if isinstance(rule, Mapping):
            rule = ModificationRule(**rule)
        for name in rule.names:
            try:
                if self.store[name] is rule:
                    continue
                else:
                    alt_rule = self.store[name]
                    if abs(alt_rule.mass - rule.mass) > 1e-3:
                        continue
                    self.store[name] += rule
            except KeyError:
                self.store[name] = rule

    def remove(self, rule):
        """Remove an existing :class:`~.ModificationRule` from this object.

        If `rule` is a :class:`~.ModificationRule`, all of its `names` will be
        removed. If `rule` is a :class:`str`, then only that name will be removed.

        Parameters
        ----------
        rule : :class:`~.ModificationRule` or :class:`str`
            Either a :class:`~.ModificationRule`, or a single :class:`str` denoting
            the name of the rule to remove.
        """
        try:
            names = rule.names
        except AttributeError:
            names = [rule]
        for name in names:
            self.store.pop(name, None)


class RestrictedModificationTable(ModificationTable):

    other_modifications = {}

    def __init__(self, rules=None, constant_modifications=None, variable_modifications=None):
        super(RestrictedModificationTable, self).__init__(rules)
        if constant_modifications is None:
            constant_modifications = []
        self.constant_modifications = constant_modifications
        if variable_modifications is None:
            variable_modifications = []
        self.variable_modifications = variable_modifications
        self.collect_available_modifications()

    def collect_available_modifications(self):
        '''Clears the rules stored in the core dictionary (on self, but not its data members),
        and copies all information in the data member sub-tables into the core dictionary'''
        modifications = {}

        for name in self.constant_modifications + self.variable_modifications:
            mod = deepcopy(self[name])
            try:
                target = extract_targets_from_string(title_cleaner.search(name).groupdict()["target"])
                mod.clear_targets()
                mod.targets = {target}
            except Exception:
                pass
            if mod.name in modifications:
                modifications[mod.name] += mod
            else:
                modifications[mod.name] = mod

        self.store.clear()
        for mod in modifications.values():
            self.add(mod)
        self._include_other_rules()


def rule_string_to_specialized_rule(rule_string: str) -> ModificationRule:
    from .modification import Modification
    name, targets = title_cleaner.search(rule_string).groups()
    targets = {extract_targets_from_string(targets)}
    # look up the rule in the main name cache, and copy it.
    rule = Modification(name).rule.clone()
    rule.targets = targets
    return rule


class ModificationMassSearch(object):
    def __init__(self, rules):
        self.index = [(rule.mass, rule) for rule in rules]
        self.index.sort(key=lambda x: (x[0], str(x[1])))

    def find_lower_bound(self, mass, error_tolerance=1e-5):
        i = bisect.bisect_left(self.index, (mass, ''))
        while i > 0:
            rmass, _rule = self.index[i]
            if error_tolerance <= abs(rmass - mass) / mass:
                i -= 1
            else:
                break
        return i

    def search_mass(self, mass, error_tolerance=1e-5):
        i = self.find_lower_bound(mass, error_tolerance)
        out = []
        for j in range(i, len(self.index)):
            rmass, rule = self.index[j]
            if abs(rmass - mass) / mass > error_tolerance:
                break
            out.append(rule)
        return out



import os
import re
import textwrap
import warnings

from io import BytesIO
from collections import OrderedDict, defaultdict
try:
    from collections.abc import Mapping, Sequence as SequenceABC
except ImportError:
    from collections import Mapping, Sequence as SequenceABC

from six import text_type

from glycopeptidepy.structure.residue import UnknownAminoAcidException, symbol_to_residue
from glycopeptidepy.structure.sequence import ProteinSequence
from glycopeptidepy.structure.parser import parse_simple
from glycopeptidepy.utils.sequence_tree import SuffixTree
from glypy.utils.base import opener

from .cv.peff import peff_cv_term


class FastaHeader(Mapping):
    """Hold parsed properties of a FASTA sequence's
    definition line.

    This object supports the :class:`Mapping` interface, and
    keys may be accessed by attribute access notation.

    Attributes
    ----------
    defline : str
        The textual representation of definition line which
        lead to the collection of attributes stored therein.
    """
    def __init__(self, mapping, original=None):
        Mapping.__init__(self)
        self._mapping = mapping
        if original is None:
            self.defline = self._make_defline()
        else:
            self.defline = original

    def __getitem__(self, key):
        return self._mapping[key]

    def __iter__(self):
        return iter(self._mapping)

    def items(self):
        return self._mapping.items()

    def keys(self):
        return self._mapping.keys()

    def values(self):
        return self._mapping.values()

    def __len__(self):
        return len(self._mapping)

    def __contains__(self, key):
        return key in self._mapping

    def __getattr__(self, key):
        if key == "_mapping":
            raise AttributeError(key)
        try:
            return self._mapping[key]
        except KeyError:
            raise AttributeError(key)

    def __str__(self):
        return self.defline

    def __repr__(self):
        return "{self.__class__.__name__}({self._mapping})".format(self=self)

    def _make_defline(self):
        raise NotImplementedError()

    def __hash__(self):
        return hash(self.defline)

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return not (self == other)

    def __dir__(self):
        base = set(dir(super(FastaHeader, self)))
        keys = set(self._mapping.keys())
        return list(base | keys)

    def startswith(self, prefix):
        '''Check if the definition line starts with `prefix`.

        If `prefix` starts with ">", trim it.

        Parameters
        ----------
        prefix: str
            The prefix string to test for

        Returns
        -------
        bool
        '''
        if self.defline is None:
            return False
        if prefix[0] == ">"  and self.defline[0] != ">":
            prefix = prefix[1:]
        return self.defline.startswith(prefix)

    def endswith(self, suffix):
        '''Check if the definition line ends with `suffix`.

        Parameters
        ----------
        suffix: str
            The suffix string to test for

        Returns
        -------
        bool
        '''
        if self.defline is None:
            return False
        return self.defline.endswith(suffix)


class UnparsableDeflineError(ValueError):
    """Indicate that a definition line could
    not be parsed by the current parser.
    """
    pass


class DefLineParserBase(object):
    """A base class for definition line parser, providing
    some common machinery in the :meth:`__call__` method.

    This class requires that the subclass provide a :meth:`parse`
    method to convert a string (without a ">" prefix). and return
    an instance of :class:`.OrderedDict`.
    """
    def __call__(self, defline):
        if defline.startswith(">"):
            defline = defline[1:]
        return self._make_header(self.parse(defline), defline)

    def _make_header(self, mapping, defline):
        return FastaHeader(mapping, defline)

    def parse(self, defline):
        raise NotImplementedError()


class PEFFDeflineParser(DefLineParserBase):
    kv_pattern = re.compile(r"\\(?P<key>\S+)=(?P<value>.+?)(?:\s(?=\\)|$)")
    detect_pattern = re.compile(r"^>?\S+:\S+")
    has_feature_index = re.compile(r"^\(?(\d+):")

    def __init__(self, validate=True):
        self.validate = validate

    def _make_header(self, mapping, defline):
        return PEFFFastaHeader(mapping, defline)

    def extract_parenthesis_list(self, text):
        chunks = []
        chunk = []
        paren_level = 0
        i = 0
        n = len(text)
        while i < n:
            c = text[i]
            i += 1
            if c == "(":
                if paren_level > 0:
                    chunk.append(c)
                paren_level += 1
            elif c == ")":
                if paren_level > 1:
                    chunk.append(c)
                paren_level -= 1
                if paren_level == 0:
                    if chunk:
                        chunks.append(chunk)
                    chunk = []
            else:
                chunk.append(c)
        chunks = list(map(''.join, chunks))
        return chunks

    def split_pipe_separated_tuple(self, text):
        parts = text.split("|")
        return parts

    def coerce_types(self, key, value):
        value = value.strip()
        feature_id_match = self.has_feature_index.search(value)
        if feature_id_match:
            feature_id = int(feature_id_match.group(1))
            value = self.has_feature_index.sub('', value)
        else:
            feature_id = None
        if "|" in value:
            value = self.split_pipe_separated_tuple(value)
            result = []
            for i, v in enumerate(value):
                result.append(self._coerce_value(key, v, i))
            return PEFFFeature(*result, feature_type=key, id=feature_id)
        else:
            return self._coerce_value(key, value, 0)

    def _coerce_value(self, key, value, index):
        try:
            return int(value)
        except ValueError:
            pass
        try:
            return float(value)
        except ValueError:
            pass
        return str(value)

    def parse(self, defline):
        if self.validate:
            match = self.detect_pattern.match(defline)
            if not match:
                raise UnparsableDeflineError(
                    "Failed to parse {!r} using {!r}".format(
                        defline, self))
        storage = OrderedDict()
        prefix = None
        db_uid = None
        if defline.startswith(">"):
            defline = defline[1:]
        try:
            prefix, defline = defline.split(":", 1)
            db_uid, defline = defline.split(" ", 1)
        except ValueError:
            raise UnparsableDeflineError(defline)
        storage['Prefix'] = prefix
        storage['Tag'] = db_uid
        kv_pattern = re.compile(r"\\(?P<key>\S+)=(?P<value>.+?)(?:\s(?=\\)|$)")
        for key, value in kv_pattern.findall(defline):
            if self.validate:
                try:
                    peff_cv_term(key, strict=True)
                except KeyError:
                    warnings.warn("PEFF Key {} not recognized".format(key))
            if not (value.startswith("(") or " (" in value):
                storage[key] = self.coerce_types(key, value)
            else:
                # multi-value
                storage[key] = [
                    self.coerce_types(key, v) for v in self.extract_parenthesis_list(value)]
        return storage


class PEFFFeature(SequenceABC):
    def __init__(self, *fields, **kwargs):
        SequenceABC.__init__(self)
        self.fields = tuple(fields)
        self.id = kwargs.get('id')
        self.feature_type = kwargs.get("feature_type")

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(str(self))

    def __getitem__(self, i):
        return self.fields[i]

    def __len__(self):
        return len(self.fields)

    def __repr__(self):
        return repr(tuple(self))

    def __str__(self):
        return "(%s%s)" % (
            '%r:' % self.id if self.id is not None else '',
            '|'.join(map(str, self)), )


class PEFFFastaHeader(FastaHeader):
    def __init__(self, mapping, original=None):
        self._feature_map = {}
        FastaHeader.__init__(self, mapping, original=original)
        self._init_feature_id_map()

    def _init_feature_id_map(self):
        for _, values in self.items():
            if isinstance(values, list):
                for feature in values:
                    if hasattr(feature, 'id'):
                        feature_id = feature.id
                        if feature_id is not None:
                            self._feature_map[feature_id] = feature
            else:
                feature = values
                if hasattr(feature, 'id'):
                    feature_id = feature.id
                    if feature_id is not None:
                        self._feature_map[feature_id] = feature

    def _make_defline(self):
        template = text_type(
            ">{self.Prefix}:{self.Tag} {rest}")
        rest_parts = []
        for key, value in self.items():
            if key in ("Prefix", "Tag"):
                continue
            part = '\\{key}={fmt_value}'
            if isinstance(value, ((int, float, text_type, str))):
                fmt_value = str(value)
            else:
                fmt_value = ''.join(map(str, value))
            rest_parts.append(part.format(key=key, fmt_value=fmt_value))
        return template.format(self=self, rest=' '.join(rest_parts))


class CallableDefLineParser(DefLineParserBase):
    """Parse a definition line using an arbitrary
    provided :class:`.Callable`.

    Attributes
    ----------
    callable_obj : Callable
    """
    def __init__(self, callable_obj):
        if not callable(callable_obj):
            raise TypeError("Must provide a callable object.")
        self.callable_obj = callable_obj

    def parse(self, defline):
        result = self.callable_obj(defline)
        if isinstance(result, Mapping):
            return OrderedDict(result.items())
        elif isinstance(result, SequenceABC):
            return OrderedDict(enumerate(result))
        else:
            raise UnparsableDeflineError(
                "Failed to parse {!r} using {!r} (return value: {!r})".format(
                    defline, self, result))


def split_on_whitespace(string):
    return re.split(r"\s+", string)


space_delim_parser = CallableDefLineParser(split_on_whitespace)


class RegexDefLineParser(DefLineParserBase):
    """Parse a definition line using a provided
    regular expression.

    If :attr:`pattern` contains named groups, this will
    produce a :class:`OrderedDict` where the group names
    are keys, otherwise their index will be used as the
    key.

    Attributes
    ----------
    pattern : re.pattern
    """
    def __init__(self, pattern):
        if isinstance(pattern, str):
            self.pattern = re.compile(pattern)
        else:
            self.pattern = pattern
        self._is_keyed = "?P<" in self.pattern.pattern

    def parse(self, defline):
        match = self.pattern.search(defline)
        if match:
            if not self._is_keyed:
                return OrderedDict(enumerate(match.groups()))
            else:
                return (match.groupdict())
        else:
            raise UnparsableDeflineError(
                "Failed to parse {!r} using {!r}".format(
                    defline, self))


class DispatchingDefLineParser(DefLineParserBase):
    """Attempt to parse a definition line using one of many
    :class:`DeflineParserBase` instances.

    Attributes
    ----------
    graceful : bool
        If :attr:`graceful`, failure to parse a defline
        with all of :attr:`parsers` results in a :class:`FastaHeader`
        no properties other than *defline*
    header_type : type
        The type to use to contain the result of parsing.
        Defaults to :class:`FastaHeader`.
    parsers : list
        A :class:`list` of :class:`DeflineParserBase` instances
    """
    def __init__(self, parsers, graceful=False, header_type=FastaHeader):
        self.parsers = list(parsers)
        self.graceful = graceful
        self.header_type = header_type

    def _make_header(self, mapping, defline):
        return self.header_type(mapping, defline)

    def parse(self, defline):
        for parser in self.parsers:
            try:
                parts = parser.parse(defline)
                return parts
            except UnparsableDeflineError:
                continue

        if self.graceful:
            return FastaHeader({}, defline)
        else:
            raise UnparsableDeflineError(
                "Failed to parse {:r} using {:r}".format(
                    defline, self))


uniprot_regex = r"(?P<db>[a-z^\|]+)\|(?P<accession>[a-zA-Z0-9\-]+)\|(?P<name>\S*)(?:\s(?P<description>.+))?"
uniprot_parser = RegexDefLineParser(uniprot_regex)

partial_uniprot_regex = r"(?P<accession>[A-Z0-9\-]+)\|(?P<name>\S+)"
partial_uniprot_parser = RegexDefLineParser(partial_uniprot_regex)

peff_parser = PEFFDeflineParser()

default_parser = DispatchingDefLineParser(
    [uniprot_parser, partial_uniprot_parser, peff_parser, space_delim_parser])


class FastaFileReader(object):
    '''A base class for parsing amino acid Fasta Files. Supports
    sequential access through iteration, and random access if :attr:`index`
    is :const:`True`.
    '''
    def __init__(self, path, defline_parser=default_parser, encoding='utf8', index=False, case_sensitive=False):
        self.state = "defline"
        self.handle = opener(path, 'rb')
        self.defline_parser = defline_parser
        self.encoding = encoding
        self.case_sensitive = case_sensitive
        self._generator = None
        self._index = None
        if index:
            self._build_index()

    def _build_index(self):
        try:
            pos = self.handle.tell()
            self.handle.seek(0)
        except (AttributeError, ValueError, TypeError) as ex:
            print(ex)
            pos = None
        indexer = FastaIndexer(encoding=self.encoding, defline_parser=self.defline_parser)
        index = indexer.build_index(self.handle)
        if pos is not None:
            self.handle.seek(pos)
        self._index = index

    @property
    def index(self):
        return self._index

    def __len__(self):
        if self.index is not None:
            return len(self.index)
        else:
            raise TypeError("Cannot get the length of an un-indexed FASTA")

    def __getitem__(self, key):
        if self.index is not None:
            try:
                pos = self.handle.tell()
                self.handle.seek(0)
            except (AttributeError, ValueError, TypeError):
                pos = None
            if isinstance(key, int):
                start, end = tuple(self.index.values())[key]
            elif isinstance(key, slice):
                return [self[i] for i in range(*key.indices(len(self.index)))]
            else:
                start, end = self.index[key]
            self.handle.seek(start)
            data = self.handle.read(end - start)
            value = next(self._parse_lines(BytesIO(data)))
            if pos is not None:
                self.handle.seek(pos)
            return value
        else:
            raise ValueError("No index was built")

    def process_result(self, d: dict):
        return d

    def __iter__(self):
        return self

    def __next__(self):
        if self._generator is None:
            if self.index is None:
                self._generator = self._parse_lines(self.handle)
            else:
                self._generator = (self[k] for k in self.index)
        return next(self._generator)

    def reset(self):
        """Restart the iterator from the beginning of the sequence list.

        Attempts to seek to the beginning of the file. If the input file
        is not seekable, this will throw an exception.
        """
        self.handle.seek(0)
        self._generator = None

    def next(self):
        return self.__next__()

    def _parse_lines(self, stream):
        sequence_pattern = re.compile(r"^(\s+|>)")
        sequence_chunks = []
        defline = None
        state = 'defline'
        for line in stream:
            line = line.strip()
            try:
                line = line.decode(self.encoding)
            except AttributeError:
                pass
            if state == 'defline':
                if line.startswith(">"):
                    defline = line[1:]
                    state = "sequence"
                else:
                    continue
            else:
                if not sequence_pattern.match(line):
                    sequence_chunks.append(line)
                else:
                    if defline is not None:
                        try:
                            value = ''.join(sequence_chunks).replace(" ", "")
                            if not self.case_sensitive:
                                value = value.upper()
                            yield self.process_result({
                                "name": self.defline_parser(defline),
                                "sequence": value
                            })
                        except KeyError as e:
                            warnings.warn(str(e))
                    sequence_chunks = []
                    defline = None
                    state = 'defline'
                    if line[0] == '>':
                        defline = re.sub(r"[\n\r]", "", line[1:])
                        state = "sequence"

        if sequence_chunks:
            try:
                value = ''.join(sequence_chunks).replace(" ", "")
                if not self.case_sensitive:
                    value = value.upper()
                yield self.process_result(
                    {
                        "name": self.defline_parser(defline),
                        "sequence": value
                    })
            except KeyError as e:
                warnings.warn(str(e))


FastaFileParser = FastaFileReader


class ProteinFastaFileReader(FastaFileReader):
    '''Parses Fasta Files into :class:`~.ProteinSequence` instances. Supports
    sequential access, as well as random access when :attr:`index` is :const:`True`.

    Attributes
    ----------
    replace_unknown: :class:`bool`
        Whether to replace amino acids which are not recognized with an 'X'
        instead of throwing an error. If not set, the parser will replace
        the amino acids and emit a warning instead.
    case_sensitive: :class:`bool`
        Whether to convert all characters in the amino acid sequence to upper case
        prior to parsing.
    '''
    def __init__(self, path, defline_parser=default_parser, encoding='utf8', index=False,
                 replace_unknown=None, case_sensitive=False):
        super(ProteinFastaFileReader, self).__init__(
            path, defline_parser, encoding, index=index, case_sensitive=case_sensitive)
        self.replace_unknown = replace_unknown
        self._replace_amino_acid_pattern = None

    def _make_replace_amino_acid_pattern(self):
        pattern = re.compile(r"[^%s]" % (
            ''.join(symbol_to_residue.keys()), ))
        return pattern

    def _replace_unknown_amino_acids(self, sequence):
        if self._replace_amino_acid_pattern is None:
            self._replace_amino_acid_pattern = self._make_replace_amino_acid_pattern()
        pattern = self._replace_amino_acid_pattern
        return pattern.sub("X", sequence)

    def process_result(self, d: dict) -> ProteinSequence:
        try:
            p = ProteinSequence(
                d['name'], d['sequence'], parser_funciton=parse_simple, annotations=dict(d['name']))
        except UnknownAminoAcidException:
            if self.replace_unknown is None or self.replace_unknown:
                if self.replace_unknown is None:
                    warnings.warn("Replacing unknown amino acids in %s" % d['name'])
                d['sequence'] = self._replace_unknown_amino_acids(d['sequence'])
                p = ProteinSequence(
                    d['name'], d['sequence'], parser_funciton=parse_simple,
                    annotations=dict(d['name']))
            else:
                raise
        return p


ProteinFastaFileParser = ProteinFastaFileReader


class FastaFileWriter(object):

    def __init__(self, handle, encoding='utf8'):
        self.handle = handle
        self.encoding = encoding
        try:
            if 'b' not in self.handle.mode:
                raise ValueError(
                    "FASTA files writer must be opened in binary mode! Make sure to open the file with 'wb'.")
        except (AttributeError, TypeError):
            pass

    def write(self, defline, sequence):
        if isinstance(defline, FastaHeader):
            defline = str(defline).encode(self.encoding)
        elif isinstance(defline, text_type):
            defline = defline.encode(self.encoding)
        if not defline.startswith(b">"):
            defline = b">" + defline
        if isinstance(sequence, text_type):
            sequence = sequence.encode(self.encoding)
        self.handle.write(defline)
        self.handle.write(b"\n")
        self.handle.write(sequence)
        self.handle.write(b"\n\n")

    def writelines(self, iterable):
        for defline, seq in iterable:
            self.write(defline, seq)


class ProteinFastaFileWriter(FastaFileWriter):
    def __init__(self, handle, encoding='utf8', line_length=80):
        super(ProteinFastaFileWriter, self).__init__(handle, encoding)
        self.line_length = line_length

    def write(self, protein): # pylint: disable=arguments-differ
        defline = str(protein.name)
        seq = '\n'.join(textwrap.wrap(protein.get_sequence(), self.line_length))
        super(ProteinFastaFileWriter, self).write(defline, seq)

    def writelines(self, iterable):
        for protein in iterable:
            self.write(protein)


def read(f, defline_parser=default_parser, reader_type=ProteinFastaFileParser, **kwargs):
    return reader_type(f, defline_parser=defline_parser, **kwargs)


class PEFFHeaderBlock(Mapping):
    def __init__(self, storage=None, block_type=None):
        Mapping.__init__(self)
        if storage is None:
            storage = OrderedDict()
        self.block_type = block_type
        self._storage = storage

    def keys(self):
        return self._storage.keys()

    def values(self):
        return self._storage.values()

    def items(self):
        return self._storage.items()

    def __iter__(self):
        return iter(self._storage)

    def __getitem__(self, key):
        return self._storage[key]

    def __len__(self):
        return len(self._storage)

    def __contains__(self, key):
        return key in self._storage

    def __getattr__(self, key):
        if key == "_storage":
            raise AttributeError(key)
        try:
            return self._storage[key]
        except KeyError:
            raise AttributeError(key)

    def __dir__(self):
        base = set(dir(super(PEFFHeaderBlock, self)))
        keys = set(self._storage.keys())
        return list(base | keys)

    def __str__(self):
        return "%s" % '\n'.join('# %s=%s' % kv for kv in self.items())

    def __repr__(self):
        return "{self.__class__.__name__}({self._storage})".format(self=self)


class PEFFReader(ProteinFastaFileReader):
    def __init__(self, path, defline_parser=PEFFDeflineParser(False), index=False):
        super(PEFFReader, self).__init__(
            opener(path, 'rb'), defline_parser, encoding='ascii', index=index)
        try:
            if 'b' not in self.handle.mode:
                raise ValueError(
                    "PEFF files must be opened in binary mode! Make sure to open the file with 'rb'.")
        except (AttributeError, TypeError):
            pass
        self.version = (0, 0)
        self.blocks = []
        self.comments = []
        self.number_of_entries = 0
        self._parse_header()

    def _parse_header(self):
        offset = 0
        line = self.handle.readline()
        offset = len(line)
        line = line.decode('ascii')
        if not line.startswith("# PEFF"):
            raise ValueError("Not a PEFF File")
        self.version = tuple(map(int, line.strip()[7:].split(".")))
        in_header = True
        current_block = defaultdict(list)
        while in_header:
            line = self.handle.readline()
            offset = len(line)
            line = line.decode('ascii')
            if not line.startswith("#"):
                in_header = False
                self.handle.seek(-offset, os.SEEK_CUR)
            line = line.strip()[2:]
            if '=' in line:
                key, value = line.split("=", 1)
                if key == "GeneralComment":
                    self.comments.append(value)
                else:
                    current_block[key].append(value)
            if line.startswith("//"):
                if current_block:
                    self.blocks.append(PEFFHeaderBlock(OrderedDict((k, v if len(v) > 1 else v[0])
                                                                   for k, v in current_block.items())))
                current_block = defaultdict(list)
        number_of_entries = 0
        for block in self.blocks:
            try:
                number_of_entries += int(block['NumberOfEntries'])
            except KeyError:
                pass
        self.number_of_entries = number_of_entries


class PEFFWriter(ProteinFastaFileWriter):
    version = (1, 0)

    def __init__(self, handle, line_length=80):
        super(PEFFWriter, self).__init__(handle, 'ascii', line_length)

    def write_header(self, blocks, comments=None):
        if comments is None:
            comments = []
        if blocks is None:
            blocks = []
        self.handle.write("# PEFF %d.%d\n" % self.version)
        for comment in comments:
            self.handle.write("# GeneralComment=%s\n" % comment)
        self.handle.write("# //\n")
        for block in blocks:
            self.handle.write(str(block))
            self.handle.write("\n# //\n")


class FastaIndexer(object):
    '''Encapsulates the process of building an index of byte-level offsets
    and entry-level metadata for a Fasta file.
    '''
    def __init__(self, read_size=1000000, encoding='utf8', defline_parser=default_parser):
        self.read_size = int(read_size)
        self.encoding = encoding
        self.defline_parser = defline_parser

    def _chunk_iterator(self, stream):
        delim = b"\n>"
        read_size = self.read_size
        f = stream
        buff = f.read(read_size)
        started_with_with_delim = buff.startswith(delim)
        parts = buff.split(delim)
        tail = parts[-1]
        front = parts[:-1]
        i = 0
        for part in front:
            i += 1
            if part == b'':
                continue
            if i == 1:
                if started_with_with_delim:
                    yield delim + part
                else:
                    yield part
            else:
                yield delim + part
        running = True
        while running:
            buff = f.read(read_size)
            if not buff:
                running = False
                buff = tail
            else:
                buff = tail + buff
            parts = buff.split(delim)
            tail = parts[-1]
            front = parts[:-1]
            for part in front:
                yield delim + part
        yield delim + tail

    def _generate_offsets(self, stream):
        i = 0
        defline_pattern = re.compile(br"\s*>([^\n\r]+)[\n\r]+")
        for line in self._chunk_iterator(stream):
            match = defline_pattern.match(line)
            if match:
                yield i, match.group(1)
            i += len(line)
        yield i, None

    def _handle_defline(self, defline):
        defline = defline.decode(self.encoding)
        if self.defline_parser:
            try:
                defline = self.defline_parser(defline)
            except UnparsableDeflineError:
                warnings.warn("Could not parse defline %r" % (defline,))
        return defline

    def build_index(self, stream):
        '''Build an random access and metadata index over the stream
        of Fasta File entries.

        Parameters
        ----------
        stream: file-like

        Returns
        -------
        :class:`FastaIndex`
        '''
        index = OrderedDict()
        g = self._generate_offsets(stream)
        last_offset = 0
        last_defline = None
        for offset, defline in g:
            if last_defline is not None:
                index[self._handle_defline(last_defline)] = (last_offset, offset)
            last_defline = defline
            last_offset = offset
        assert last_defline is None
        return FastaIndex(index)


class FastaIndex(object):
    '''A :class:`~.Mapping`-like object which stores the byte offsets
    and entry metadata for a Fasta file.
    '''
    def __init__(self, mapping):
        self.mapping = mapping
        self._suffix = None
        self._index_map = dict()

    def __repr__(self):
        template = "{self.__class__.__name__}({size:d})"
        return template.format(self=self, size=len(self))

    def __getitem__(self, key):
        return self.mapping[key]

    def __iter__(self):
        return iter(self.mapping)

    def __contains__(self, key):
        return key in self.mapping

    def keys(self):
        return self.mapping.keys()

    def values(self):
        return self.mapping.values()

    def items(self):
        return self.mapping.items()

    def __len__(self):
        return len(self.mapping)

    def _build_suffix_tree(self):
        sfx_tree = SuffixTree()
        for k in self:
            sfx_tree.add_ngram(str(k), k)
        return sfx_tree

    def suffix(self, key):
        '''Search for headers where ``key`` is a suffix of
        the header.

        This method builds a suffix tree which may be very
        expensive in memory

        Parameters
        ----------
        key: :class:`str`
            The suffix to search for

        Returns
        -------
        :class:`list`
        '''
        if self._suffix is None:
            self._suffix = self._build_suffix_tree()
        return self._suffix.subsequences_of(key)

    def _build_index_for_key(self, key):
        index = defaultdict(set)
        for header in self.keys():
            try:
                value = header[key]
                index[value].add(header)
            except KeyError:
                continue
        self._index_map[key] = index

    def index_by(self, key):
        '''Build an extra index over the values of ``key``
        '''
        self._build_index_for_key(key)

    def clear_indices(self):
        '''Discard all the extra indices.
        '''
        self._index_map.clear()
        self._suffix = None

    def query(self, queries=None, **kwargs):
        '''Query the index for headers whose keys match the values passed
        through ``kwargs``

        Parameters
        ----------
        queries: :class:`dict`, optional
            A mapping of :class:`FastaHeader` keys to values to match for
            inclusion in the result set. All pairs must match for a header
            to be included in the results.
        **kwargs:
            Additional key-value pairs to search for

        Results
        -------
        :class:`list`:
            The :class:`FastaHeader` instances that matched all of the constraints
            the query.
        '''
        if not queries:
            queries = dict()
        else:
            queries = dict(queries)
        queries.update(kwargs)
        queries = queries.items()
        # The indices fully cover this query, so make it fast.
        if set(kwargs) <= set(self._index_map):
            index_matches = []
            for k, v in queries:
                try:
                    index_matches.append(self._index_map[k][v])
                except KeyError:
                    continue
            if not index_matches:
                return []
            return list(index_matches[0].intersection(*index_matches[1:]))
        else:
            # Employ partial index coverage to shrink the set of elements to
            # explore exhaustively
            index_keys = set(kwargs) & set(self._index_map)
            if index_keys:
                query_set = self.query({k: queries[k] for k in index_keys})
                for k in index_keys:
                    queries.pop(k)
            else:
                # No helpful index, so iterate over all headers.
                query_set = self.keys()

            matches = []
            # Check each entry for key-value equality, keeping those
            # cases which match on *all* keys. This isn't SQL, so we
            # cannot express complex relationships or conditionals
            for header in query_set:
                for k, v in queries:
                    try:
                        if header[k] != v:
                            break
                    except KeyError:
                        break
                else:
                    matches.append(header)
        return matches

    def prefix_filter(self, prefix):
        return [entry for entry in self if entry.startswith(prefix)]

    def suffix_filter(self, suffix):
        return [entry for entry in self if entry.endswith(suffix)]



from glycopeptidepy.utils.collectiontools import _AccumulatorBag

from .base import FragmentationStrategyBase

from .glycan import (
    StubGlycopeptideStrategy,
    GlycanCompositionFragmentStrategyBase,
    CADFragmentationStrategy,
    _MonosaccharideDefinitionCacher,
    OxoniumIonStrategy)

from .peptide import (
    PeptideFragmentationStrategyBase,
    HCDFragmentationStrategy,
    EXDFragmentationStrategy,
    exd_sidechain_losses)

__all__ = [
    "FragmentationStrategyBase",
    "_AccumulatorBag",

    "StubGlycopeptideStrategy", "GlycanCompositionFragmentStrategyBase",
    "CADFragmentationStrategy", "_MonosaccharideDefinitionCacher",
    "OxoniumIonStrategy",

    "PeptideFragmentationStrategyBase",
    "HCDFragmentationStrategy", "EXDFragmentationStrategy",
    "exd_sidechain_losses"
]
